/*
	+--------------------------------------+
	| 3c90xC RTLinux NetWork Device Driver |
	+--------------------------------------+

(C) 1999-2001
    Navigators Group - www.navigators.di.fc.ul.pt

       Pedro Martins        Antonio Casimiro    Paulo Verissimo     
    pmartins@di.fc.ul.pt    casim@di.fc.ul.pt   pjv@di.fc.ul.pt    

--------------------------------------------------------------- 		




 (*) This driver is based (in some parts a port) on Donald Becker 3c59x.c Linux
     Driver.
*/


#ifndef MODULE
#define MODULE
#endif


#include <linux/module.h>
#include <rtl_sync.h>
#include <rtl_core.h>
#include <linux/net.h>
#include <linux/netdevice.h> 
#include "vortex_defs.h"
#include "TCBnetdev.h"
#include "../autoconf.h"

//For Test Only
char evtcode=1;
unsigned long enviadas=0;
unsigned long recebidas_nav05=0;
unsigned long orecebidas=0;


int vortex_debug=0;

char netdev[]=CONFIG_3COM_IFACE;   //network device used - included from ../autoconf.h

#define ibuffers 32
#define obuffers 16
#define PKT_BUF_SZ 1536
unsigned long macADDR;
unsigned char started=0;  

TCBnetdev* TCBdev;

void* rx_backup[ibuffers];

// Linux Device
struct device* rtdev=0;

extern struct device* root_vortex_dev;

#define BUFFER_SIZE_TO_FDUPLEX 10

void* input_buffers[ibuffers+BUFFER_SIZE_TO_FDUPLEX];
void* output_buffers[obuffers];
unsigned int bufflen[ibuffers+BUFFER_SIZE_TO_FDUPLEX];
hrtime_t     msgrcvtime[ibuffers+BUFFER_SIZE_TO_FDUPLEX]; 	

/* Pre-Cyclone chips have no documented multicast filter, so the only
   multicast setting is to receive all multicast frames.  At least
   the chip has a very clean way to set the mode, unlike many others. */
static void rtset_rx_mode(struct device *dev)
{
	long ioaddr = dev->base_addr;
	int new_mode;

	if (dev->flags & IFF_PROMISC) {
		if (vortex_debug > 0)
			printk("%s: Setting promiscuous mode.\n", dev->name);
		new_mode = SetRxFilter|RxStation|RxMulticast|RxBroadcast|RxProm;
	} else	if ((dev->mc_list)  ||  (dev->flags & IFF_ALLMULTI)) {
		new_mode = SetRxFilter|RxStation|RxMulticast|RxBroadcast;
	} else
		new_mode = SetRxFilter | RxStation | RxBroadcast;
	
	outw(new_mode, ioaddr + EL3_CMD);
}


/*
 * Handle uncommon interrupt sources.  This is a separate routine to minimize
 * the cache impact.
 */
static void
rtvortex_error(struct device *dev, int status)
{
	struct vortex_private *vp = (struct vortex_private *)dev->priv;
	long ioaddr = dev->base_addr;
	int do_tx_reset = 0;
	int i;

	if (status & TxComplete) {			/* Really "TxError" for us. */
		unsigned char tx_status = inb(ioaddr + TxStatus);
		/* Presumably a tx-timeout. We must merely re-enable. */
		if (vortex_debug > 2
			|| (tx_status != 0x88 && vortex_debug > 0))
			printk("%s: Transmit error, Tx status register %2.2x.\n",
				   dev->name, tx_status);
		if (tx_status & 0x14)  vp->stats.tx_fifo_errors++;
		if (tx_status & 0x38)  vp->stats.tx_aborted_errors++;
		outb(0, ioaddr + TxStatus);
		if (tx_status & 0x30)
			do_tx_reset = 1;
		else					/* Merely re-enable the transmitter. */
			outw(TxEnable, ioaddr + EL3_CMD);
	}
	if (status & RxEarly) {				/* Rx early is unused. */
		rtboomerang_rx(dev);   
		outw(AckIntr | RxEarly, ioaddr + EL3_CMD);
	}
	if (status & StatsFull) {			/* Empty statistics. */
			EL3WINDOW(5);
			outw(SetIntrEnb | (inw(ioaddr + 10) & ~StatsFull), ioaddr + EL3_CMD);
			EL3WINDOW(7);		
	}
	if (status & IntReq) {		/* Restore all interrupt sources.  */
		outw(vp->status_enable, ioaddr + EL3_CMD);
		outw(vp->intr_enable, ioaddr + EL3_CMD);
	}
	if (status & HostError) {
		u16 fifo_diag;
		EL3WINDOW(4);
		fifo_diag = inw(ioaddr + Wn4_FIFODiag);
		if (vortex_debug > 0)
			printk("%s: Host error, FIFO diagnostic register %4.4x.\n",
				   dev->name, fifo_diag);
		/* Adapter failure requires Tx/Rx reset and reinit. */
		if (vp->full_bus_master_tx) {
			outw(TotalReset | 0xff, ioaddr + EL3_CMD);
			for (i = 2000; i >= 0 ; i--)
				if ( ! (inw(ioaddr + EL3_STATUS) & CmdInProgress))
					break;
			/* Re-enable the receiver. */
			outw(RxEnable, ioaddr + EL3_CMD);
			outw(TxEnable, ioaddr + EL3_CMD);
		} else if (fifo_diag & 0x0400)
			do_tx_reset = 1;
		if (fifo_diag & 0x3000) {
			outw(RxReset, ioaddr + EL3_CMD);
			for (i = 2000; i >= 0 ; i--)
				if ( ! (inw(ioaddr + EL3_STATUS) & CmdInProgress))
					break;
			/* Set the Rx filter to the current state. */
			rtset_rx_mode(dev);
			outw(RxEnable, ioaddr + EL3_CMD); /* Re-enable the receiver. */
			outw(AckIntr | HostError, ioaddr + EL3_CMD);
		}
	}
	if (do_tx_reset) {
		int j;
		outw(TxReset, ioaddr + EL3_CMD);
		for (j = 200; j >= 0 ; j--)
			if ( ! (inw(ioaddr + EL3_STATUS) & CmdInProgress))
				break;
		outw(TxEnable, ioaddr + EL3_CMD);
	}

}

static void rtvortex_tx_timeout(struct device *dev)      
{
	struct vortex_private *vp = (struct vortex_private *)dev->priv;
	long ioaddr = dev->base_addr;
	int j;
	
	printk("%s: transmit timed out, tx_status %2.2x status %4.4x.\n",
		   dev->name, inb(ioaddr + TxStatus),
		   inw(ioaddr + EL3_STATUS));
		   
	if ((inb(ioaddr + TxStatus) & 0x88) == 0x88)
		printk("%s: Transmitter encountered 16 collisions --"
			   " network cable problem?\n", dev->name);
	if (inw(ioaddr + EL3_STATUS) & IntLatch) {
		printk("%s: Interrupt posted but not delivered --"
			   " IRQ blocked by another device?\n", dev->name);
			   
		rtcard_interrupt(dev->irq,0);
	}
	outw(TxReset, ioaddr + EL3_CMD);
	for (j = 200; j >= 0 ; j--)
		if ( ! (inw(ioaddr + EL3_STATUS) & CmdInProgress))
			break;

	if (vp->full_bus_master_tx) {
		int i;
		printk("  Flags; bus-master %d, full %d; dirty %d "
			   "current %d.\n",
			   vp->full_bus_master_tx, vp->tx_full, vp->dirty_tx, vp->cur_tx);
		printk("  Transmit list %8.8x vs. %p.\n",
			   inl(ioaddr + DownListPtr),
			   &vp->tx_ring[vp->dirty_tx % obuffers]);
		for (i = 0; i < obuffers; i++) {
			printk("  %d: @%p  length %8.8x status %8.8x\n", i,
				   &vp->tx_ring[i],
				   le32_to_cpu(vp->tx_ring[i].length),
				   le32_to_cpu(vp->tx_ring[i].status));
		}
	}

	if (vp->full_bus_master_tx) {
		if (vortex_debug > 0)
			printk("%s: Resetting the Tx ring pointer.\n",
				   dev->name);
		if (vp->cur_tx - vp->dirty_tx > 0  &&  inl(ioaddr + DownListPtr) == 0)
			outl(virt_to_bus(&vp->tx_ring[vp->dirty_tx % obuffers]),
				 ioaddr + DownListPtr);
		if (vp->tx_full && (vp->cur_tx - vp->dirty_tx <= obuffers - 1)) {
			vp->tx_full = 0;
			clear_bit(0, (void*)&dev->tbusy);
		}
		outb(PKT_BUF_SZ>>8, ioaddr + TxFreeThreshold);
		outw(DownUnstall, ioaddr + EL3_CMD);
	} 
	
	/* Issue Tx Enable */
	outw(TxEnable, ioaddr + EL3_CMD);
	dev->trans_start = jiffies;
	
	/* Switch to register set 7 for normal use. */
	EL3WINDOW(7);
}

static int getMTU(void)
{
	return mtu;
}

static unsigned long getMacAddr(void)
{
	return macADDR; 
}

static void startReceive(void)
{
	started=1;
}

static void stopReceive(void)
{
	started=0;
}

static int
send(void* buff,int len,unsigned char count)  
{
	struct device* dev=rtdev;
	struct vortex_private *vp = (struct vortex_private *)dev->priv;
	long ioaddr = dev->base_addr;
	rtl_irqstate_t state;
	int a;
	
	
	if (test_and_set_bit(0, (void*)&dev->tbusy) != 0) {
		if (jiffies - dev->trans_start >= TX_TIMEOUT)
			rtvortex_tx_timeout(dev);
		return -1;
	} else {
	 	int entry,entry2,entryf ;
		struct boom_tx_desc *prev_entry;
		unsigned long flags;
		int i;
		
		rtl_no_interrupts(state);

//For Latency Evaluation
#ifdef MACHINE_SEND
	outb(evtcode,0x378);
	enviadas++;	
	(*(unsigned char*)buff)=evtcode;
	evtcode++;
	if (evtcode==16)
	 evtcode=1;
#endif		
		
		/* Calculate the next Tx descriptor entry. */
		entry = vp->cur_tx % obuffers;
		
		prev_entry =
			&vp->tx_ring[(vp->cur_tx-1) % obuffers];

		if (vortex_debug > 3)
			printk("%s: Trying to send a packet, Tx index %d.\n",
				   dev->name, vp->cur_tx);
		if ((vp->tx_full) || ( vp->cur_tx+count+1 - vp->dirty_tx > obuffers - 1 ) ) {
			if (vortex_debug >0)
				printk("%s: Tx Ring full, refusing to send buffer.\n",
					   dev->name);
			rtl_restore_interrupts(state);		   
			return -1;
		}
		
		entryf = (vp->cur_tx+count-1) % obuffers;				
		
		len+=14;
		
		memcpy((output_buffers[entryf]+14),buff,len);
		
		for (a=0;a<count;a++)
		{
		 entry2 = vp->cur_tx % obuffers;				
		 vp->tx_ring[entry2].next = virt_to_bus(&vp->tx_ring[(vp->cur_tx+1) % obuffers]);
		 vp->tx_ring[entry2].addr = virt_to_bus(output_buffers[entryf]);
		 vp->tx_ring[entry2].length = cpu_to_le32(len|LAST_FRAG);
		 vp->tx_ring[entry2].status = cpu_to_le32(len|TxIntrUploaded);  //len |...
		 vp->cur_tx++;
		}
		vp->tx_ring[entry2].next = 0;
				
		outw(DownStall, ioaddr + EL3_CMD);
		/* Wait for the stall to complete. */
		for (i = 600; i >= 0 ; i--)
			if ( (inw(ioaddr + EL3_STATUS) & CmdInProgress) == 0)
				break;
		prev_entry->next = cpu_to_le32(virt_to_bus(&vp->tx_ring[entry]));
		if (inl(ioaddr + DownListPtr) == 0) {
			outl(virt_to_bus(&vp->tx_ring[entry]), ioaddr + DownListPtr);
		}
		outw(DownUnstall, ioaddr + EL3_CMD);		
		//vp->cur_tx++;
		if (vp->cur_tx - vp->dirty_tx > obuffers - 1)
			vp->tx_full = 1;
		else {					/* Clear previous interrupt enable. */
			prev_entry->status &= cpu_to_le32(~TxIntrUploaded);
			clear_bit(0, (void*)&dev->tbusy);
		}
		dev->trans_start = jiffies;
		for(i=0;i<count;i++)
		 self_rx(buff,len);	//simulate a fullduplex send
		rtl_restore_interrupts(state);
		return 0;
	}
}


int read(void* addr,hrtime_t* rvcTime)
{
	int entry;
	struct vortex_private *vp = (struct vortex_private *)rtdev->priv;			
	long ioaddr = rtdev->base_addr;
	rtl_irqstate_t state;
	int i,tam;
	
	rtl_no_interrupts(state);
	
	while (vp->cur_rx>vp->dirty_rx)
	{
		outw(UpStall, ioaddr + EL3_CMD);		
		/* Wait for the stall to complete. */
		for (i = 600; i >= 0 ; i--)
			if ( (inw(ioaddr + EL3_STATUS) & CmdInProgress) == 0)
				break;
		entry = vp->dirty_rx % ibuffers;
		if (*((unsigned short*)(input_buffers[entry]+12))==ntohs(RT_PROT_ID))
		{
			memcpy(addr,(input_buffers[entry]+14),bufflen[entry]);
		}
		vp->rx_ring[entry].status = 0;	/* Clear complete bit. */
		vp->dirty_rx++;
		outw(UpUnstall, ioaddr + EL3_CMD);		
		
		if (*((unsigned short*)(input_buffers[entry]+12))==ntohs(RT_PROT_ID))
		{
			rtl_restore_interrupts(state);	
			*rvcTime=msgrcvtime[entry];
			return bufflen[entry];		
		}	
	}
		
	for (i=ibuffers;i<ibuffers+BUFFER_SIZE_TO_FDUPLEX;i++)
	{
		if (bufflen[i]!=0)
		{
			memcpy(addr,input_buffers[i],bufflen[i]);			
			*rvcTime=msgrcvtime[i];
			tam=bufflen[i];
			bufflen[i]=0;
			rtl_restore_interrupts(state);
			return tam;
		}
	}
		
	rtl_restore_interrupts(state);
	return 0;
}



static void rtcard_interrupt(int irq,struct pt_regs *regs)   
{
	struct device *dev = rtdev;
	struct vortex_private *vp = (struct vortex_private *)dev->priv;
	long ioaddr;
	int latency, status;
	int work_done = max_interrupt_work;

#if defined(__i386__)
	/* A lock to prevent simultaneous entry bug on Intel SMP machines. */
	if (test_and_set_bit(0, (void*)&dev->interrupt)) {
		printk(KERN_ERR"%s: SMP simultaneous entry of an interrupt handler.\n",
			   dev->name);
		dev->interrupt = 0;	/* Avoid halting machine. */
		return;
	}
#else
	if (dev->interrupt) {
		printk(KERN_ERR "%s: Re-entering the interrupt handler.\n", dev->name);
		return;
	}
	dev->interrupt = 1;
#endif

	dev->interrupt = 1;
	ioaddr = dev->base_addr;
	latency = inb(ioaddr + Timer);
	status = inw(ioaddr + EL3_STATUS);
	
	if (vortex_debug > 4)
		printk("%s: interrupt, status %4.4x, latency %d ticks.\n",
			   dev->name, status, latency);
	do {
		if (vortex_debug > 5)
				printk("%s: In interrupt loop, status %4.4x.\n",
					   dev->name, status);
		if (status & UpComplete) {
			outw(AckIntr | UpComplete, ioaddr + EL3_CMD);			
			rtboomerang_rx(dev);
		}

		if (status & TxAvailable) {
			if (vortex_debug > 5)
				printk( "	TX room bit was handled.\n");
			// There's room in the FIFO for a full-sized packet. 
			outw(AckIntr | TxAvailable, ioaddr + EL3_CMD);
			clear_bit(0, (void*)&dev->tbusy);
		}

		if (status & DownComplete) {
			unsigned int dirty_tx = vp->dirty_tx;

			while (vp->cur_tx - dirty_tx > 0) {
				int entry = dirty_tx % obuffers;
				if (inl(ioaddr + DownListPtr) ==
					virt_to_bus(&vp->tx_ring[entry]))
					break;			// It still hasn't been processed. 
					
				vp->stats.tx_packets++; // Counted below. 
				dirty_tx++;
			}
			vp->dirty_tx = dirty_tx;
			outw(AckIntr | DownComplete, ioaddr + EL3_CMD);
			if (vp->tx_full && (vp->cur_tx - dirty_tx <= obuffers - 1)) {
				vp->tx_full= 0;
				clear_bit(0, (void*)&dev->tbusy);
			}
		}
		if (status & DMADone) {
			if (inw(ioaddr + Wn7_MasterStatus) & 0x1000) {
				outw(0x1000, ioaddr + Wn7_MasterStatus); // Ack the event. 
				 
				if (inw(ioaddr + TxFree) > 1536) {
					clear_bit(0, (void*)&dev->tbusy);
				} else // Interrupt when FIFO has room for max-sized packet. 
					outw(SetTxThreshold + (1536>>2), ioaddr + EL3_CMD);
			}
		}
		// Check for all uncommon interrupts at once. 
		if (status & (HostError | RxEarly | StatsFull | TxComplete | IntReq)) {
			if (status == 0xffff)
				break;
			rtvortex_error(dev, status);
		}

		if (--work_done < 0) {
			if ((status & (0x7fe - (UpComplete | DownComplete))) == 0) {
				// Just ack these and return. 
				outw(AckIntr | UpComplete | DownComplete, ioaddr + EL3_CMD);
			} else {  
			   //ver melhor!!!!!
			   printk("3COMnetdrv - fatal error - too much work in interrupt\n");
			   break;
			}
		}
		// Acknowledge the IRQ. 
		outw(AckIntr | IntReq | IntLatch, ioaddr + EL3_CMD);
		if (vp->cb_fn_base)			
			writel(0x8000, vp->cb_fn_base + 4);

	} while ((status = inw(ioaddr + EL3_STATUS)) & (IntLatch | RxComplete));

	if (vortex_debug > 4)
		printk( "%s: exiting interrupt, status %4.4x.\n",
			   dev->name, status);

#if defined(__i386__)
	clear_bit(0, (void*)&dev->interrupt);
#else
	dev->interrupt = 0;
#endif
	rtl_hard_enable_irq(dev->irq);
	return;
}


static void reinit_input_buffers(void)
{
	int i;
	struct vortex_private *vp = (struct vortex_private *)rtdev->priv;
	long ioaddr = rtdev->base_addr;

        outw(UpStall, ioaddr + EL3_CMD);
        /* Wait for the stall to complete. */
		for (i = 600; i >= 0 ; i--)
			if ( (inw(ioaddr + EL3_STATUS) & CmdInProgress) == 0)
				break;
	//printk("3COM RTL Driver - Input Buffers Full - reiniting the buffers\n");
	
	vp->dirty_rx=0;vp->cur_rx=0;
	for (i = 0; i < ibuffers; i++) 		 
		vp->rx_ring[i].status = 0;	/* Clear complete bit. */
						
	outl(virt_to_bus(&vp->rx_ring[0]), rtdev->base_addr + UpListPtr);
	outw(UpUnstall, ioaddr + EL3_CMD);
	for (i=ibuffers;i<ibuffers+BUFFER_SIZE_TO_FDUPLEX;i++)
	 bufflen[i]=0;
}

static int
self_rx(void* buff,int len)             
{
	int i;
	
	for (i=ibuffers;i<ibuffers+BUFFER_SIZE_TO_FDUPLEX;i++)
	if (bufflen[i]==0)
	{
		bufflen[i]=len;
		memcpy(input_buffers[i],buff,len);
		msgrcvtime[i]=gethrtime();
		return 1;
	}
	
	return 0;
}


static int
rtboomerang_rx(struct device *dev)             //OK
{
	struct vortex_private *vp = (struct vortex_private *)dev->priv;
	int entry = vp->cur_rx % ibuffers,entry2;
	long ioaddr = dev->base_addr;
	int rx_status;
	int rx_work_limit = vp->dirty_rx + ibuffers - vp->cur_rx;
	int i;
	
	if (vortex_debug > 5)
		printk("  In boomerang_rx(), status %4.4x, rx_status "
			   "%4.4x.\n",
			   inw(ioaddr+EL3_STATUS), inw(ioaddr+RxStatus));
	if (--rx_work_limit < 0)
			{
				reinit_input_buffers();	
			}	
	else		   
	while ((rx_status = le32_to_cpu(vp->rx_ring[entry].status)) & RxDComplete){
	        //free all buffers
		if (--rx_work_limit < 0)
			{
			 reinit_input_buffers();	
			 break;
			}
		if (rx_status & RxDError) { /* Error, update stats. */
			unsigned char rx_error = rx_status >> 16;
			if (vortex_debug > 2)
				printk(" Rx error: status %2.2x.\n", rx_error);
			vp->stats.rx_errors++;
			if (rx_error & 0x01)  vp->stats.rx_over_errors++;
			if (rx_error & 0x02)  vp->stats.rx_length_errors++;
			if (rx_error & 0x04)  vp->stats.rx_frame_errors++;
			if (rx_error & 0x08)  vp->stats.rx_crc_errors++;
			if (rx_error & 0x10)  vp->stats.rx_length_errors++;
		} else {
			/* packet length */
			int pkt_len = rx_status & 0x1fff;
			
			
// EVALUATION	

#ifdef MACHINE_SEND		
	// verificar se e pacote de pause
	if
	  (*((unsigned char*)(input_buffers[entry]+0))!=0xFF)	
	  {
	   printk("PAUSE\n");
	  }
	 
	//ver se vem da nav05  
	if (*((unsigned char*)(input_buffers[entry]+11))==0x53)
	{
	 recebidas_nav05++;
	}  
	 else
	  orecebidas++;
	  
	  
#endif
			
			vp->stats.rx_bytes += pkt_len;
			
			bufflen[entry]=pkt_len - 14;			
			msgrcvtime[entry]=gethrtime();
					
			dev->last_rx = jiffies;
			vp->stats.rx_packets++;
//#ifdef MACHINE_SEND
                if (started==0)    //ignore the packet
                {
	         outw(UpStall, ioaddr + EL3_CMD);		
	         /* Wait for the stall to complete. */
		for (i = 600; i >= 0 ; i--)
			if ( (inw(ioaddr + EL3_STATUS) & CmdInProgress) == 0)
				break;
		 entry2 = vp->dirty_rx % ibuffers;
		 vp->rx_ring[entry2].status = 0;	/* Clear complete bit. */
		 vp->dirty_rx++;
		 outw(UpUnstall, ioaddr + EL3_CMD);		
		} 
//#endif        
		}
		entry = (++vp->cur_rx) % ibuffers;
	}
	
	
	return 0;
}

void about(void)
{
	printk("\n");
	printk("+------------------------------+\n");
	printk("|    3Com - 3c900,3c905 V.1    |\n");
	printk("|      TCB Network Driver      |\n");
	printk("+------------------------------+\n");
	printk("(C) 2000 - Navigators Group\n");
	printk("           Faculty of Sciences, University of Lisbon\n");
	printk("           PORTUGAL\n\n");
}



int init_buffers()
{	
	int i,a;	
	struct vortex_private *vp = (struct vortex_private *)rtdev->priv;
	long ioaddr = rtdev->base_addr;
	
	for (i=ibuffers;i<ibuffers+BUFFER_SIZE_TO_FDUPLEX;i++)
	{
	 bufflen[i]=0;
	 input_buffers[i]=kmalloc(PKT_BUF_SZ,0);
	 if (input_buffers[i] == 0)
		 {
		   printk("Failed alloc network card buffer\n");
		   for (a=ibuffers;a<i;a++)
		     kfree(input_buffers[a]);
		   return -1;
		 }
	 
	} 
	
	outw(UpStall, ioaddr + EL3_CMD);		
	         /* Wait for the stall to complete. */
		for (i = 600; i >= 0 ; i--)
			if ( (inw(ioaddr + EL3_STATUS) & CmdInProgress) == 0)
				break;		
	
	if (vp->full_bus_master_rx) { /* Boomerang bus master. */
	 vp->cur_rx = vp->dirty_rx = 0;
	for (i = 0; i < ibuffers; i++) {
		input_buffers[i]=kmalloc(PKT_BUF_SZ,0);
		if (input_buffers[i] == 0)
		 {
		   printk("Failed alloc network card buffer\n");
		   for (a=0;a<i;a++)
		     kfree(input_buffers[a]);
		   return -1;
		 }
		 
		vp->rx_ring[i].next = cpu_to_le32(virt_to_bus(&vp->rx_ring[i+1]));
		vp->rx_ring[i].status = 0;	/* Clear complete bit. */
		vp->rx_ring[i].length = cpu_to_le32(PKT_BUF_SZ | LAST_FRAG);	
		rx_backup[i]=vp->rx_ring[i].addr;
		vp->rx_ring[i].addr = cpu_to_le32(virt_to_bus(input_buffers[i]));
		}
		
		vp->rx_ring[i-1].next = cpu_to_le32(virt_to_bus(&vp->rx_ring[0]));												
						
		outl(virt_to_bus(&vp->rx_ring[0]), rtdev->base_addr + UpListPtr);				
	}
	
	if (vp->full_bus_master_tx) { 		/* Boomerang bus master Tx. */
		vp->cur_tx = vp->dirty_tx = 0;
		outb(PKT_BUF_SZ>>8, rtdev->base_addr + TxFreeThreshold); /* Room for a packet. */
		/* Clear the Tx ring. */
		outl(0, rtdev->base_addr + DownListPtr);
	}
	
	outw(UpUnstall, ioaddr + EL3_CMD);	
	
	for (i=0; i<obuffers; i++)
	{
		output_buffers[i]=kmalloc(PKT_BUF_SZ,0);
		if (output_buffers[i] == 0)
		 {
		   printk("Failed alloc network card buffer\n");
		   for (a=0;a<i;a++)
		     kfree(output_buffers[a]);
		   return -1;
		 }
		
		//ethernet destination address - broadcast
		for (a=0;a<6;a++)
			*((unsigned char*)(output_buffers[i]+a))=0xff;
		
			
		//ethernet source adress
		for (a = 6; a < 12; a++)
		  *((unsigned char*)(output_buffers[i]+a))=rtdev->dev_addr[a-6];				  
		
		//packet type
		*((unsigned short*)(output_buffers[i]+12))=htons(RT_PROT_ID);  
					
	}
	
	return 0;
}

char isBoomerang(int chip_id)
{
  if (
    (chip_id == 5) ||
    (chip_id == 6) ||
    (chip_id == 7) ||
    (chip_id == 8) ||
    (chip_id == 9) ||
    (chip_id == 11) ||
    (chip_id == 12) ||
    (chip_id == 13) ||
    (chip_id == 14) ||
    (chip_id == 22) 
    )
    return 1;
    
  return 0;
}


int init_drv(void)
{
	
	struct device* linuxdevs=dev_base;
	int a;
	int drv=0;
	struct vortex_private *vp;
	rtl_irqstate_t irqstate;
	int i;
	
	printk("searching linux %s (3c59x) device...\n",netdev);
	while (linuxdevs!=0)
	   {
	   	/* find linux 3c59x device driver*/
	   	if ((strcmp(linuxdevs->name,netdev)==0)/*&&((root_vortex_dev==linuxdevs)||(root_vortex_dev->next==linuxdevs) )*/)
	   	 {
	   	   drv=1;		 
	   	   vp = (struct vortex_private *)linuxdevs->priv;  	      	 
	   	   if (isBoomerang(vp->chip_id))
	   	   {
	   	     printk("%s -> OK\n",pci_tbl[vp->chip_id].name);	
	   	     rtdev=linuxdevs;
	   	     break;
	   	   }  
	   	 }
	    	linuxdevs=linuxdevs->next;
	   }      
	   
	   
	if (linuxdevs==0) 
	{
	 if (drv==0)
	  printk("Cannot find any linux 3com net device driver on %s\n",netdev);
	   else
	    printk("Cannot find any 3c900 or 3c905 netcard on %s\n",netdev);
	 return -1;
	}
	
	
	printk("ethernet card adress = ");
	for (a = 0; a < 6; a++)
		  printk("%X:",rtdev->dev_addr[a]);
	
	macADDR= (
	          (rtdev->dev_addr[0] <<40)+
	          (rtdev->dev_addr[1] <<32)+
	          (rtdev->dev_addr[2] <<24)+
	          (rtdev->dev_addr[3] <<16)+
	          (rtdev->dev_addr[4] <<8)+
	          (rtdev->dev_addr[5]) 
	         ); 
	         
	printk("\n");	  				  	
	printk("removing linux interrupt handler\n");
	rtl_no_interrupts(irqstate);		
	free_irq(rtdev->irq, rtdev);
	printk("removing linux card timer\n");
	del_timer(&vp->timer);

	printk("initing the network drive internal buffers <send: %u, receive: %u>\n",obuffers,ibuffers);
	if (init_buffers())
	{
	 rtl_restore_interrupts(irqstate); 	
	 return -1; 
	} 
	printk("setting rtlinux network card interrupt handler\n");
	
#ifdef MACHINE_SEND
	
	// set window 3
	outw(SelectWindow +3, rtdev->base_addr + EL3_CMD);
	
	outw(0x0020, rtdev->base_addr + Wn3_MAC_Ctrl);		
	outw(SetRxFilter|RxProm, rtdev->base_addr + EL3_CMD);
	printk("%X\n",inw(rtdev->base_addr + Wn3_MAC_Ctrl));
#endif	
	
	if (rtl_request_irq(rtdev->irq,rtcard_interrupt)!=0)
	 {
	   printk("cannot set rtlinux interrupt handler\n");
	   rtl_restore_interrupts(irqstate); 	
	   return -1;
	 }
	rtl_restore_interrupts(irqstate); 	
	rtl_hard_enable_irq(rtdev->irq);
		
	
		
	TCBdev=registerTCBnetdevice("3Com 3c900,3c905 - Navigators FCUL",&__this_module);  
	TCBdev->send=&send;
	TCBdev->read=&read;
	TCBdev->getMTU=&getMTU;
	TCBdev->getMacAddr=&getMacAddr;
	TCBdev->startReceive=&startReceive;
	TCBdev->stopReceive=&stopReceive;
	
	
	return 0;
}


int init_module(void)
{
	about();
	
	return init_drv();	
}


void cleanup_module(void)
{	
	int i;
	rtl_irqstate_t irqstate;
	struct vortex_private *vp=rtdev->priv;
	
	rtl_no_interrupts(irqstate);		
	rtl_free_irq(rtdev->irq);	
	
	outl(0, rtdev->base_addr + UpListPtr);	
			
	for (i=0;i<ibuffers;i++)
	{
		vp->rx_ring[i].addr=rx_backup[i];		
		kfree(input_buffers[i]);
	}
	
	for (i=ibuffers;i<ibuffers+BUFFER_SIZE_TO_FDUPLEX;i++)
	{
		kfree(input_buffers[i]);
	}
	
	for (i=0;i<obuffers;i++)
	{
		kfree(output_buffers[i]);
	}
	
	rtl_restore_interrupts(irqstate);
	
	//printk("Enviadas: %u\n",enviadas);
	//printk(" Recebidas da Nav05 %u\n",recebidas_nav05);
	//printk(" Recebidas de outras %u\n",orecebidas);
	
}


