/*
	(TCB)
	Distributed Failure Detection Demo Application

--[TCB 0.1]---------------------------------------------------

(C) 1999-2001
    Navigators Group - www.navigators.di.fc.ul.pt

       Pedro Martins        Antonio Casimiro    Paulo Verissimo     
    pmartins@di.fc.ul.pt    casim@di.fc.ul.pt   pjv@di.fc.ul.pt    

--------------------------------------------------------------- 			
*/	

#include "tcb.h"
#include <signal.h>
#include "defs.h"
#include <sched.h>

char child_stack[1000];
char child_stack_end;
TCBsocketDesc socketDesc;

int thrd1;

char p[4324];

int tcbDesc,wait_info_desc;
char ip_addr[3000];

void exit_on_error(char * msg,int code)
{
	if (errno==0)
	 printTCBerror(code);
	  else
	   perror(msg);
	   
	kill(thrd1,SIGINT);   
	TCBsocketClose(&socketDesc,wait_info_desc);   
	closeTCBConnection(tcbDesc,wait_info_desc);   
	exit(0);   
}

void sigint_handler(int par)
{
	
	kill(thrd1,SIGINT);    	
	TCBsocketClose(&socketDesc,wait_info_desc);   
	closeTCBConnection(tcbDesc,wait_info_desc);   
	printf("\nTCBPing stopped by user\n\n");
	exit(0);
}

void display_fail_info(fail_info_to_app* fi)
{
	int c;
	int mult=1;
	
	printf("********************************************\n");
	printf("Message from: %u, Seq: %u\n",fi->from,fi->seq);
	printf("********************************************\n");
	
	for (c=0;c<NUMBER_OF_MACHINES;c++)
	 {
	   if (fi->failure_info[c].fail!=2)
	   {
	     if (fi->failure_info[c].fail==1)
	      {
	        printf("Timing failure on the reception (machine %u)\n",c); 
	        mult=1;
	      }
	       else
	        {
	        	printf("Timely reception on machine %u\n",c); 
	        	mult=-1;
	        }
	        
	    if (fi->failure_info[c].lateness_avail==1)
	      {
	        printf("lateness degree=%d\n",fi->failure_info[c].lateness_degree*mult); 
	      }
	       else
	        {
	        	printf("Lateness degree not available\n",c); 
	        }    	
	     printf("\n");            
	   }	   
	 }
	 
}

int fn(void* arg)
{
	fail_info_to_app fi;
	
	while (1)
	{
	  TCBwaitInfo(wait_info_desc,&fi);  	
  	  display_fail_info(&fi);
	}
	return 0;
}


int main(int argc,char* argv[])
{
	int res;
	
	int ret;	
	TCBtimestamp sndtime,rcv_evt;
	unsigned int tag;
	TCBtimeinterval delay,error;	
	Func_Id p;
 	struct sockaddr_in cnt,serv;
 	int len; 	
 	char buffer[1000];
	void **data;
	unsigned long long cnts;
	pingResp* pt;
	MSGheader* pt2;
	

	printf("\n----------------------------------------------------------\n");
	printf("Distributed failure detection service - client aplication\n");
	printf("----------------------------------------------------------\n\n");
	
	        	
	if ((ret=openTCBConnection(&tcbDesc,&wait_info_desc))<OK)
	{
		printTCBerror(ret);
		exit(1);
	}
	
	//create a thread to read the WaitInfo data  	
	thrd1=clone(&fn,&child_stack_end,CLONE_VM,(void*)0);
	
	signal(SIGINT,sigint_handler);
	
	if ((ret=TCBsocket(AF_INET,&socketDesc,tcbDesc))<0)
	{
	       exit_on_error("Error on creating a TCB socket...\n\n",ret);
	}
	
 	bzero((char*)&cnt,0,sizeof(cnt));
	
 	cnt.sin_family=AF_INET;
 	cnt.sin_addr.s_addr=htonl(INADDR_ANY);
 	cnt.sin_port=htons(0);
 	
 	if ((ret=TCBbind((struct sockaddr*)&cnt,sizeof(cnt),&socketDesc))<0)
 	 {	
 	 	exit_on_error("Error on binding a TCB socket...\n",ret);
 	 }
 	
 	
 	bzero((char*)&serv,0,sizeof(serv));
 	
 	serv.sin_family=AF_INET;
 	serv.sin_addr.s_addr=inet_addr("255.255.255.255");
 	serv.sin_port=htons(16006);
 	
 	cnts=0;
 	
 	// handler for timely timing failure detection
 	strcpy(p.module,"app_module"); 
 	strcpy(p.funcname,"timelyExec"); 
 	strcpy(p.wcetname,"timelyExec_wcet"); 
 	
 	
 	while (1)
 	{
 	cnts++;
 	if ((cnts>400000))  
  	{
 	 sndtime=getTCBTimestamp(tcbDesc);	
 																 
  	 if ((ret=TCBsendtoWFD((char*)buffer,20,0,(struct sockaddr*)&serv,sizeof(serv),DESTINATION_MASK,16006,&socketDesc,sndtime,400,2000000,p))<0)
   	  {	
  	  	if ((ret!=ALREADY_FAILED)&&(ret!=BURSTY_SENDS)&&(ret!=TOO_MANY_DISTRIBUTED_FD))													
  	  	exit_on_error("On sending a packet in a TCB socket...\n",ret);
  	  }  	  
  	 cnts=0;
  	}
 
 	len=sizeof(cnt);
 	if ((ret=TCBrecvfrom((char*)buffer,sizeof(pingResp),0,(struct sockaddr*)&cnt,&len,&socketDesc,&rcv_evt,&delay,&error,NON_BLOCKING))>0) 	
 	{ 	
 	 pt=(pingResp*)(buffer+sizeof(MSGheader));	
 	 printf("Packet From %s:\n",inet_ntoa(cnt.sin_addr.s_addr));
 	 printf("------------------------------------\n");
 	 printf("Send   : time <%llu> us error <%llu> us SEQ:%u\n",pt->delay/1000,pt->error/1000,pt->seq);
 	 printf("\n");
 	}	  
 	 else
 	 if (ret<0)   // ret==0 means no info on NON_BLOCKING read
 	  {
 	     exit_on_error("On getting a message from a TCB socket...\n",ret);
 	  }
 	}
}

