/*
	(TCB)
	Distributed Failure Detection Demo Application

--[TCB 0.1]---------------------------------------------------

(C) 1999-2001
    Navigators Group - www.navigators.di.fc.ul.pt

       Pedro Martins        Antonio Casimiro    Paulo Verissimo     
    pmartins@di.fc.ul.pt    casim@di.fc.ul.pt   pjv@di.fc.ul.pt    

--------------------------------------------------------------- 			
*/

#include "tcb.h"
#include <signal.h>
#include "defs.h"
#include <sched.h>

char child_stack[1000];
char child_stack_end;

char p[4324];

int thrd1;

int tcbDesc,wait_info_desc;
TCBsocketDesc socketDesc;
void exit_on_error(char * msg,int code)
{
	if (errno==0)
	 printTCBerror(code);
	  else
	   perror(msg);
	kill(thrd1,SIGINT);      
	TCBsocketClose(&socketDesc,wait_info_desc);      
	closeTCBConnection(tcbDesc,wait_info_desc);   
	exit(1);   
}

void sigint_handler(int par)
{
	kill(thrd1,SIGINT);   
	TCBsocketClose(&socketDesc,wait_info_desc);   
	closeTCBConnection(tcbDesc,wait_info_desc);   
	printf("\nTCBPing Server stopped by user\n\n");
	exit(0);
}


void display_fail_info(fail_info_to_app* fi)
{
	int c;
	int mult=1;
	
	printf("********************************************\n");
	printf("Message from: %u, Seq: %u\n",fi->from,fi->seq);
	printf("********************************************\n");
	
	for (c=0;c<NUMBER_OF_MACHINES;c++)
	 {
	   if (fi->failure_info[c].fail!=2)
	   {
	     if (fi->failure_info[c].fail==1)
	      {
	        printf("Timing failure on the reception (machine %u)\n",c); 
	        mult=1;
	      }
	       else
	        {
	        	printf("Timely reception on machine %u\n",c); 
	        	mult=-1;
	        }
	        
	    if (fi->failure_info[c].lateness_avail==1)
	      {
	        printf("lateness degree=%d\n",fi->failure_info[c].lateness_degree*mult); 
	      }
	       else
	        {
	        	printf("Lateness degree not available\n",c); 
	        }    	
	     printf("\n");            
	   }	   
	 }
	 
}

int fn(void* arg)
{
fail_info_to_app fi;
	
	while (1)
	{
	TCBwaitInfo(wait_info_desc,&fi);  	
  	display_fail_info(&fi);
	}
	return 0;
}

int main()
{
 	int res;
	
	int ret;
	char ip_addr[300];
	TCBtimestamp sndtime,rcv_evt;
	unsigned int tag;
	TCBtimeinterval delay,error;	
 	struct sockaddr_in serv,cnt;
 	int len; 	
 	char buffer[1000];
	void **data;
	pingResp* pt;

	
	printf("\n---------------------------------------------------------\n");
	printf("Distributed failure detection service - server aplication\n");
	printf("----------------------------------------------------------\n\n");

	
	if ( (ret=openTCBConnection(&tcbDesc,&wait_info_desc))<OK)
	{
		printTCBerror(ret);
		exit(1);
	}
	
	thrd1=clone(&fn,&child_stack_end,CLONE_VM,(void*)0);
	
	signal(SIGINT,sigint_handler);
	
	if ((ret=TCBsocket(AF_INET,&socketDesc,tcbDesc))<0)
	{
		   exit_on_error("Error on creating a TCB socket...\n\n",ret);
	}
	
 	bzero((char*)&serv,0,sizeof(cnt));
 	
 	serv.sin_family=AF_INET;
 	serv.sin_addr.s_addr=htonl(INADDR_ANY);
 	serv.sin_port=htons(16006);
 	
 	
 	if ((ret=TCBbind((struct sockaddr*)&serv,sizeof(serv),&socketDesc))<0)
 	 {	
 	 	exit_on_error("Error on binding a TCB socket...\n",ret);
 	 } 	  	  	
 	
 	//we want to get winfo about the traffic on this port 
 	if ((ret=TCBwinfoport(16006,&socketDesc))<0)
 	  exit_on_error("Error on spec winfoport...\n",ret);
 	
 	
	while (1)
	{ 	
	
 	len=sizeof(serv); 	
 	
 	if ((ret=TCBrecvfrom((char*)buffer,0,0,(struct sockaddr*)&cnt,&len,&socketDesc,&rcv_evt,&delay,&error,BLOCKING))>0) 	
 	{
 	  sndtime=getTCBTimestamp(tcbDesc);	
 	  
 	  pt=(pingResp*)(buffer+sizeof(MSGheader));	
 	  pt->error=error;
 	  pt->delay=delay;	
 	  pt->seq=((MSGheader*)buffer)->seq; 
 	 
													 	 	
  	  if ((ret=TCBsendto((char*)buffer,sizeof(pingResp),0,(struct sockaddr*)&cnt,len,&socketDesc,sndtime))<0)
  	   {
  	   	exit_on_error("On send a packet in a TCB socket...\n",ret);
  	   }
 	}
 	else
 	  exit_on_error("On getting a message from a TCB socket...\n",ret);
 	 	
	} 	
}

