/*


	+----------------------------------------+
	| LoopBack TCB NetWork Device Driver     |
	+----------------------------------------+

(C) 1999-2001
    Navigators Group - www.navigators.di.fc.ul.pt

       Pedro Martins        Antonio Casimiro    Paulo Verissimo     
    pmartins@di.fc.ul.pt    casim@di.fc.ul.pt   pjv@di.fc.ul.pt    

--------------------------------------------------------------- 

*/


#ifndef MODULE
#define MODULE
#endif


#include <linux/module.h>
#include <rtl_sync.h>
#include <rtl_core.h>
#include <linux/net.h>
#include <linux/netdevice.h> 
#include "TCBnetdev.h"


#define ibuffers 32
#define obuffers 16
#define PKT_BUF_SZ 1536
unsigned char started=0;  //application says when want to start to receive packets

TCBnetdev* TCBdev;

#define BUFFER_SIZE_TO_FDUPLEX 10

void* input_buffers[ibuffers+BUFFER_SIZE_TO_FDUPLEX];
void* output_buffers[obuffers];
unsigned int bufflen[ibuffers+BUFFER_SIZE_TO_FDUPLEX];
hrtime_t     msgrcvtime[ibuffers+BUFFER_SIZE_TO_FDUPLEX]; 	

static int getMTU(void)
{
	return 1500;
}

static void startReceive(void)
{
	started=1;
}

static void stopReceive(void)
{
	started=0;
}


// LOOPBACK Functions ***********************************

static unsigned long getMacAddr_lo(void)
{
	return 0; 
}

static int
send_lo(void* buff,int len,unsigned char count) 
{
	int i;

	for(i=0;i<count;i++)
	   self_rx(buff,len);	//simulate a fullduplex send
	   
	return 0;   
} 

int read_lo(void* addr,hrtime_t* rvcTime)
{
	int i,tam;

	for (i=ibuffers;i<ibuffers+BUFFER_SIZE_TO_FDUPLEX;i++)
	{
		if (bufflen[i]!=0)
		{
			memcpy(addr,input_buffers[i],bufflen[i]);			
			*rvcTime=msgrcvtime[i];
			tam=bufflen[i];
			bufflen[i]=0;
			return tam;
		}
	}
	return 0;	
}


static int
self_rx(void* buff,int len)             
{
	int i;
	
	for (i=ibuffers;i<ibuffers+BUFFER_SIZE_TO_FDUPLEX;i++)
	if (bufflen[i]==0)
	{
		bufflen[i]=len;
		memcpy(input_buffers[i],buff,len);
		msgrcvtime[i]=gethrtime();
		return 1;
	}
	
	return 0;
}


void about(void)
{
	printk("\n");
	printk("+------------------------------+\n");
	printk("|        LoopBack Device       |\n");
	printk("|      TCB Network Driver      |\n");
	printk("+------------------------------+\n");
	printk("(C) 2001 - Navigators Group\n");
	printk("           Faculty of Sciences, University of Lisbon\n");
	printk("           PORTUGAL\n\n");
}

int init_buffers_lo()
{	
	int i,a;	
	
	for (i=ibuffers;i<ibuffers+BUFFER_SIZE_TO_FDUPLEX;i++)
	{
	 bufflen[i]=0;
	 input_buffers[i]=kmalloc(PKT_BUF_SZ,0);
	 if (input_buffers[i] == 0)
		 {
		   printk("Failed alloc network card buffer\n");
		   for (a=ibuffers;a<i;a++)
		     kfree(input_buffers[a]);
		   return -1;
		 }
	 
	} 
	
	return 0;
}

int init_drv(void)
{
	
	
	if (init_buffers_lo())
	  return -1;	
	
	TCBdev=registerTCBnetdevice("loopback device - Navigators FCUL",&__this_module);  
	TCBdev->send=&send_lo;
	TCBdev->read=&read_lo;
	TCBdev->getMTU=&getMTU;
	TCBdev->getMacAddr=&getMacAddr_lo;
	TCBdev->startReceive=&startReceive;
	TCBdev->stopReceive=&stopReceive;
	
	return 0;
}


int init_module(void)
{
	about();
	
	return init_drv();	
}


void cleanup_module(void)
{	
	int i;
	
	
		for (i=0;i<ibuffers+BUFFER_SIZE_TO_FDUPLEX;i++)
		  kfree(input_buffers[i]);
	
}


