
#include <rtl.h>
#include <time.h>
#include <rtl_time.h>
#include "../system_defs.h"

/*
+--------------------------------+
| ERROS DEVOLVIDOS AS APLICACOES | 
+--------------------------------+
*/

#define	OK			 0
#define TCB_ACESS_ERROR		-1 	// Erro no acesso a tcb         
#define NO_AVAIL_CONN		-2      // Conexao indisponivel        
#define NO_AVAIL_LOCAL_MEASURE  -3      // medicao local nao disponivel 
#define INVALID_LOCALM_DESC	-4      // descriptor de medicao local invalido 
#define NO_AVAIL_EXEC		-5      // execucao nao disponivel 
#define FUNC_NOT_FOUND		-6      // funcao nao encontrada   
#define WCET_FUNC_NOT_FOUND	-7      // funcao wcet nao encontrada
#define CANT_RUN_TIMELY		-8      // nao consegue correr atempadamente 
#define NO_AVAIL_LOCAL_FD	-9      // detector de falhas local nao disponivel 
#define ALREADY_FAILED          -10     // a tcb detectou que a falha deu-se antes de 
				        // processar o pedido 		  	 
#define INVALID_LOCALFD_DESC	-11     // descriptor de deteccao de falhas locais
#define TOO_MANY_DISTRIBUTED_FD  -12     // demasiadas deteccoes simultaneas
#define BURSTY_SENDS		-13	// demasiados sends no round
#define IN_INIT_STATE		-14	// a tcb enconta-se em fase de inicializacao

/*
+--------------------------------------------+
| SERVICOS FORNECIDOS PELA TCB AS APLICACOES |
+--------------------------------------------+
*/

typedef 
  enum {
          get_connection, 	        // pedido de conexao a tcb             	
          end_connection,		// final de conexao
          get_timestamp,		// obter estampilha temporal
          get_id,			// obter id da TCB local
          start_localM,			// iniciar medicao local
          end_localM,			// terminar medicao local	
          exec,				// execucao local 
          start_localFD,		// iniciar detecao de falha local
          end_localFD,			// terminar deteccao de falha local
          start_distributedFD,		// iniciar detecao de falhas distribuida
          msg_received_on_payload,	// recepcao de mensagem no canal de PayLoad
          winfo_to_fifo,		// especifica para que fifo deve ser enviada a informacao
          				// relacionada com um dado porto de comunicacoes
          stop_winfo_port		// para de enviar winfo respectiva a um porto dado
       } 
Tservice;



/*
+----------------------------------------+
| CABECALHO DAS MENSAGENS ENVIADAS A TCB |
+----------------------------------------+
*/

typedef struct  {
	Tservice 	srvc;		// identificacao do servico (Tservice)
	int		sizeofmsg;      // tamanho dos parametros do servico
} ReqTcbHeader;


/*
+----------------------+
| FORMATO DAS DURACOES |
+----------------------+
*/

typedef long long TCBtimeinterval; 



/*
+---------------------+
| ESTAMPILHA TEMPORAL |
+---------------------+
*/

typedef long long TCBtimestamp;


/*
+----------------------------------------------------------------------------+
| INFORMACAO DE RETORNO DO SERVICO 'end_localM'- passada pela TCB a aplicacao|
|      						 (bibl.TCB) no fifo          |
+----------------------------------------------------------------------------+
*/

typedef struct 
        {
          char 		  stat;         // ok ou codigo erro 
          TCBtimeinterval tm;		// duracao 
          TCBtimestamp    ts;		// instante final da medicao
        }
endLocalMResult;        



/*
+-----------------------------------------------------------------------------+
| IDENTIFICACAO DE UMA FUNCAO DEFINIDA NO NUCLEO DO LINUX (funcoes que definem|
| handlers invocados aquando de falhas temporais,funcoes de execucao atempada)|
| Estrutura usada pelas aplicacoes ao aceder aos servicos fornecidos pela     |      
| bliblioteca TCB.							      |									       
+-----------------------------------------------------------------------------+
*/

typedef struct {
	char  module[50];                  // nome do modulo onde esta definida a funcao
	char  funcname[50];		// nome da funcao a executar
	char  wcetname[50];		// nome da funcao que especifica o WCET
}  
Func_Id;                                



/*
+----------------------------------------------------------------------------+
|  PARAMETROS DO SERVICO 'exec' - passados pela aplicacao (biblioteca TCB) a |
|                                 TCB num fifo.				     |
+----------------------------------------------------------------------------+
*/


typedef struct 
	{
	  TCBtimestamp	  start_ev;     // inicio temporal da execucao
	  TCBtimeinterval delay;        // offset da execucao a partir de start_ev
	  TCBtimeinterval max_exec;     // tempo maximo para o fim da execucao a partir de start_ev
	  char 		  modName[50];  // nome do modulo onde esta definida a funcao
	  char		  funcName[50]; // nome da funcao a executar
 	  char 		  wcetName[50]; // nome da funcao que especifica o WCET
 	  char 		  TraceName[20]; // identificacao da funcao para tracing
	}
execPars;



/*
+-----------------------------------------------------------------------------+
| PARAMETROS DO SERVICO 'start_localFD'-passados pela aplicacao(biblioteca TCB)|
|			               a TCB num fifo                         |
+-----------------------------------------------------------------------------+
*/

typedef struct 
	{
	  TCBtimestamp	  start_ev;     // inicio temporal da deteccao
	  TCBtimeinterval spec;         // especificacao do intervalo temporal 
	  				// para a geracao de falha
	  TCBtimeinterval deadline;     // especificacao do deadline 				
	  char 		  modName[50];  // nome do modulo onde esta definida a funcao
	  char		  funcName[50]; // nome da funcao a executar
 	  char 		  wcetName[50]; // nome da funcao que especifica o WCET
	}
localFDPars;


/*
+------------------------------------------------------------------------------------+
| PARAMETROS DO SERVICO 'start_distributedFD'-passados pela aplicacao(biblioteca TCB)|
|			               a TCB num fifo                                |
+------------------------------------------------------------------------------------+
*/

typedef struct 
	{
	  TCBtimestamp	  start_ev;     // inicio temporal da deteccao
	  TCBtimeinterval spec;         // especificacao do intervalo temporal 
	  				// para a geracao de falha
	  TCBtimeinterval deadline;     // especificacao do deadline 								
	  unsigned char	  to;						
	  unsigned int	  destination_port;
	  char 		  modName[50];  // nome do modulo onde esta definida a funcao
	  char		  funcName[50]; // nome da funcao a executar
 	  char 		  wcetName[50]; // nome da funcao que especifica o WCET
	}
distributedFDPars;	



/*
+-----------------------------------------------------------------------------+
| INFORMACAO DE RETORNO DO SERVICO 'end_localFD'- passada pela TCB a aplicacao|
|      						 (bibl.TCB) no fifo           |
+-----------------------------------------------------------------------------+
*/

typedef struct
	{
	  int             stat;	        // ok ou codigo erro 
	  TCBtimestamp    end_ev;       // instante da desactivacao da deteccao
	  TCBtimeinterval procdelay;    // intervalo temporal
	  char		  faulty;       // houve ou nao falha 
	}
endLocalFDResult;



typedef struct
	{
		unsigned long seq;
		unsigned long long time;
		unsigned char from;
	}
msg_rcv_on_payloadPars;	

typedef struct 
	{
		unsigned char fail;   // =2 means no info
		unsigned char lateness_avail;	
		unsigned long lateness_degree;
	}
fail_info;

typedef struct
	{
		unsigned  char broadcast;
		unsigned  char from;
		unsigned  long seq;
		fail_info failure_info[NUMBER_OF_MACHINES];
	}
fail_info_to_app;		



