/*	
--[TCB 0.1]---------------------------------------------------

(C) 1999-2001
    Navigators Group - www.navigators.di.fc.ul.pt

       Pedro Martins        Antonio Casimiro    Paulo Verissimo     
    pmartins@di.fc.ul.pt    casim@di.fc.ul.pt   pjv@di.fc.ul.pt    

---------------------------------------------------------------
*/


#ifndef MODULE
#define MODULE
#endif


#include <linux/module.h>
#include <rtl_sync.h>
#include <asm/io.h>

#include "defs.h"
#include "trace_defs.h"
#include <pthread.h>


// para debug
void printTimeStamp(hrtime_t ts)
{
// rtl_printf("(%X)",ts);  
 //rtl_printf("(%X)",ts<<32);
}

/*
+--------------------+
|  VARIAVEIS GLOBAIS |********************************************************
+--------------------+ 
*/

#define TRACE_TCB 0

#ifdef RNDTRIP_DEBUG
	char rnd_debug_code=1;
#endif

#ifdef BENCH
	unsigned long bench_max_exec_read=0,bench_max_exec_send=0;
	unsigned long max_roundtrip=0;
	unsigned long long next_release_of_send;
	unsigned long long next_release_of_read;
	unsigned long max_release_time_send=0;
	unsigned long max_release_time_read=0;
	unsigned long max_release_delay_of_handler_DFD=0; 
	unsigned long max_release_delay_of_handler_LFD=0; 
#endif

unsigned long   seq_numbers=1;    // 23 bits

char 		usedFifos
		[MAX_TCB_CONNECTIONS];   //tabela que contem a informacao sobre 
				         //a utilizacao dos fifos(e.g.conexoes)				         
hrtime_t	localMeasures
                [MAX_LOCAL_MEASURES];    //tabela de medicoes locais

                 
execInfo 	executions
                [MAX_EXEC];              //tabela de execucoes atempadas

localFDInfo 	localFD
                [MAX_LOCAL_FD];          //tabela de deteccoes de falhas locais

/*
+------------+
| NETWORKING |
+------------+       
*/

#define  MAX_MTU	1500

unsigned char   net_msg_memory[MAX_MTU]; // network message memory
pthread_t	snd_task_id,rcv_task_id;
unsigned char   rcv_msg_memory[MAX_MTU*(NUMBER_OF_MACHINES)];
msg_info_type	last_rcv_info[NUMBER_OF_MACHINES];    // informacao das ultimas mensagens recebidas de cada no
char 		system_view[NUMBER_OF_MACHINES]; // 0-NOT_PRESENT, 1-IN_INIT, 2-NORMAL
unsigned char   number_of_active_machines=0;
char 		msgfrom_in_this_read_slot[NUMBER_OF_MACHINES];
char            wait_round; 
char 		state_of_operation; // init, normal ...
unsigned long   read_loops=0;
unsigned char   tcb_machine_id;
unsigned char   contention_rounds;
unsigned char   my_mask;
unsigned int	netMTU;
hrtime_t	start_of_system_tasks;
unsigned long	trace_msg_number=0;

/*
+-------------------------------+
| DISTRIBUTED FAILURE DETECTION |
+-------------------------------+       
*/

int 		 number_of_detections[NUMBER_OF_MACHINES];
snd_info	 send_info[MAX_SEND_TABLE_SIZE];
rcv_info	 receive_info[MAX_RCV_TABLE_SIZE];
rcv_payload_info receive_payload_info[MAX_RCV_PLOAD_TABLE_SIZE];
wait_info	 wait_infos[MAX_WAIT_INFO_TABLE_SIZE];
int		 max_fields,max_items;
int 		 n_sends_in_this_slot;
int   	         port_to_fifo[65536];	//arranjar uma solucao + elegante!


/*
+----------------------+
|  FUNCOES AUXILIARES  |******************************************************
+----------------------+ 
*/

/*
+-----------------------------------------+
| Retorna o numero de maquinas no sistema |
| que estao no estado normal.		  |
+-----------------------------------------+
*/

inline unsigned char get_number_of_normal_state_machines()
{
	unsigned char res=0;
	int a;
	
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	{
		if (system_view[a]==NORMAL)
		  res++;
	}
	
	return res;
}

/*
+-------------------+
|  SERVICOS DA TCB  |**********************************************************
+-------------------+ 
		

+----------------------------------------+
| enviar estampilha temporal a aplicacao |
+----------------------------------------+
*/

void TCBget_timestamp(int fifo)
{
	hrtime_t t;
#ifdef RNDTRIP_DEBUG
	rtl_irqstate_t irqstate;			
	
	rtl_no_interrupts(irqstate);		   	
	outb(rnd_debug_code,0x378);
	rnd_debug_code++;
	if (rnd_debug_code==3)
	 rnd_debug_code=1;
#endif
	t=gethrtime();
#ifdef RNDTRIP_DEBUG
	rtl_restore_interrupts(irqstate);			   		
#endif
	rtf_put(fifo,&t,sizeof(hrtime_t));		
}

/*
+----------------------------------------+
| obter o identificador da TCB (nodo)	 |
+----------------------------------------+
*/

unsigned char get_tcb_id(void)
{
	unsigned long macAddr=getMacAddr();        // obter endereco ethernet 
						   // da placa de rede do canal
						   // sincrono	   
	
	switch (macAddr) {
	#include "../rt_net_addrs.h"
	}
	return 0;
}


void* failTask(void * par)   
{
	struct sched_param p;
	FD_pars     *fdp=(FD_pars*)par;
	handtype    funcao = fdp->funcao;
	unsigned long long tof=fdp->time_of_failure;
#ifdef BENCH
	unsigned long long t;
	long long diff;
#endif
	
	p.sched_priority=TCB_USER_TASK_PRIORITY;    		
   	pthread_setschedparam(pthread_self(),SCHED_FIFO,&p);

	tcb_init_user_task(pthread_self(),fdp->release_time,fdp->deadline,"s");
	
	pthread_make_periodic_np(pthread_self(),0,0xFFFFFFFFFFFFFFF);
	
	pthread_suspend_np(pthread_self());
	pthread_wait_np();
#ifdef BENCH
	t=gethrtime();
	diff=t-tof;
	if (diff>max_release_delay_of_handler_DFD)
		max_release_delay_of_handler_DFD=diff;
#endif	
	funcao();
	pthread_delete_np(pthread_self());
		
	while (1)
	{		
	}
}


/*
+----------------------------------------+
| iniciar deteccao de falha distribuida	 |
+----------------------------------------+
*/

void TCBstart_distributedFD(int fifo,distributedFDPars* pars)  
{
	long t;
	int a,c,ret,sndPt;
	char to_many=0;
	rtl_irqstate_t irqstate;			
	char to_array[NUMBER_OF_MACHINES];		
	unsigned char mask=1;
	unsigned char dest=pars->to;
	unsigned char number_of_dest=0;
	handtype     funcptr;
	wcetFunctype wcetptr;
	pthread_t    rt_task_id;
	FD_pars      fdp;
	
	if (n_sends_in_this_slot==max_items)
	{	
		t=BURSTY_SENDS;
	        rtf_put(fifo,&t,sizeof(unsigned long));		
	        return;				
	}
	
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	{
		if (mask&dest)
		{
		  to_array[a]=1;
		  number_of_dest++;
		}  
		   else
		    to_array[a]=0;
		mask<<=1;
	}
	
	if (strcmp(pars->modName,"")!=0)
	{
	
	//obter endereco da funcao	
	funcptr=get_module_symbol(pars->modName,pars->funcName);
	if (funcptr == 0)
	{
		t=FUNC_NOT_FOUND;
		rtf_put(fifo,&t,sizeof(t));		
		return;
	}
		
	wcetptr=get_module_symbol(pars->modName,pars->wcetName);
		
	if (wcetptr == 0)
	{
		t=WCET_FUNC_NOT_FOUND;
		rtf_put(fifo,&t,sizeof(t));		
		return;
	}
	}
	 else
	  funcptr=0;
	
	//sincronizar o acesso com a tarefa de leitura
	//introduz uma latencia controlada na execucao da tarefa de leitura
	//rtl_no_interrupts(irqstate);		   	
	TCBenter_in_critical_section();
	
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	  if (to_array[a]==1)
	   if (number_of_detections[a]+1>max_items)	
	   {
	    to_many=1;
	    break;
	   } 
	
	TCBleave_critical_section();	
	//rtl_restore_interrupts(irqstate);			   		
	
	if (to_many)
	{
	 t=TOO_MANY_DISTRIBUTED_FD;
//printk("TOO MANY\n");	 
	 rtf_put(fifo,&t,sizeof(unsigned long));		
	 return;
	} 
	
	
	t=0;
	TCBenter_in_critical_section();
	//rtl_no_interrupts(irqstate);		   	
			
	n_sends_in_this_slot++;
			
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	{
	  if (to_array[a])
	   number_of_detections[a]++;
	 }  
	
	//find a free entry on the send_info table		  
	for (a=0;a<MAX_SEND_TABLE_SIZE;a++)
	{
		if (send_info[a].is_free)
		 {
			t=1;
			send_info[a].is_free=0;
			send_info[a].time_of_generation=pars->start_ev;
			send_info[a].spec=pars->spec;
			send_info[a].destination_port=pars->destination_port;
			send_info[a].seq=seq_numbers;
			send_info[a].to=pars->to;
			sndPt=a;
			break;	 		
		 }
	}
	
	if (t!=0)
	{
	t=0;
	// iniciar a estrutura do WaitInfo
	for (a=0;a<MAX_WAIT_INFO_TABLE_SIZE;a++)
	{
		if (wait_infos[a].is_free)
		 {
			t=1;
			if (funcptr!=0)
			{
			 fdp.funcao=funcptr;
			 fdp.release_time=0;  //por agora nao e necessario
			 fdp.deadline=pars->deadline;
			 
			rtl_no_interrupts(irqstate);		   	 
			fdp.time_of_failure=(pars->spec*1000)+pars->start_ev;
			
			if (fdp.time_of_failure<gethrtime())
			{
				send_info[sndPt].is_free=1;	       	          	
				rtl_restore_interrupts(irqstate);
				TCBleave_critical_section();
				t=ALREADY_FAILED;
	       	          	rtf_put(fifo,&t,sizeof(t));	       	          		       	          	
	       	          	return;
				
			}
			//rtl_restore_interrupts(irqstate);
			ret=pthread_create (&rt_task_id, NULL, failTask, (void *)&fdp);
			if (ret!=0)																			
	 		 {
	 		  send_info[sndPt].is_free=1;
	       	          rtl_restore_interrupts(irqstate);
	       	          TCBleave_critical_section();
	    		  t=NO_AVAIL_EXEC;
	       	          rtf_put(fifo,&t,sizeof(t));	       	          	       	          
	       	          return;
		         }
		        }
		         else
		          rt_task_id=0;  
		          
			wait_infos[a].is_free=0;
			wait_infos[a].owner=fifo;
			wait_infos[a].failure_at_time=fdp.time_of_failure;
			wait_infos[a].send_fifo=fifo+INC_TO_WAIT_FIFO;
			wait_infos[a].seq=seq_numbers;
			wait_infos[a].to=pars->to;
			wait_infos[a].any_fail=0;
		 	wait_infos[a].n_dests=number_of_dest;
		 	wait_infos[a].rt_task_id=rt_task_id;
		 	wait_infos[a].from=tcb_machine_id;
		 	wait_infos[a].destination_port=pars->destination_port;   
		 	for (c=0;c<NUMBER_OF_MACHINES;c++)
		 	 wait_infos[a].failure_info[c].fail=2;
		 	rtl_restore_interrupts(irqstate); 
			break;	 		
		 }
	}
	}
	
	//rtl_restore_interrupts(irqstate);			   					
	TCBleave_critical_section();
	
	
	if (t==0)
	 {
	 	printk("Error: no free send_info or wait_info\n");
	 	while (1){};  //never happens if the system is correct
	 }
	
	t=seq_numbers;

	rtf_put(fifo,&t,sizeof(unsigned long));		
	
	seq_numbers++;
	
	if (seq_numbers==8388608)  //23 bits bound
	 seq_numbers=1;
	
}

/*
+------------------------------+
| enviar ID da TCB a aplicacao |
+------------------------------+
*/

void TCBget_id(int fifo)
{	
	int t=(int)get_tcb_id();

	rtf_put(fifo,&t,sizeof(t));		
}


/*
+-----------------------------------------------------------+
| inicia medicao local e devolve um descriptor a aplicacao. |
| retorna NO_AVAIL_LOCAL_MEASURE se nao existir descriptor  |
| disponivel.				       	            |
+-----------------------------------------------------------+
*/
 
void TCBstart_local_measure(int fifo,hrtime_t *initTS)  //PARA SAIR
{
	int a,desc=NO_AVAIL_LOCAL_MEASURE;
	
	for (a=0;((a<MAX_LOCAL_MEASURES)&&(desc==NO_AVAIL_LOCAL_MEASURE));a++)
	{
		if (localMeasures[a]==0)
		{
			desc=a;
			localMeasures[a]=*initTS;
		}
	}
	
	rtf_put(fifo,&desc,sizeof(int));
}


/*
+----------------------------------------------------------------+
| termina medicao local; liberta descriptor; devolve a aplicacao |
| a medida da duracao e a estampilha final                       |  
| retorna INVALID_LOCALM_DESC se o descriptor for invalido       | 
+----------------------------------------------------------------+
*/

void TCBend_local_measure(int fifo,int* descriptor)   //PARA SAIR
{
	endLocalMResult res;
	hrtime_t actTime;
	
	/* descriptor invalido ??? */
	if ( (*descriptor<0) || (*descriptor>=MAX_LOCAL_MEASURES) || (localMeasures[*descriptor]==0) ) 
	{
		res.stat=INVALID_LOCALM_DESC;
		rtf_put(fifo,&res,sizeof(res));
		return;
	}
	
	actTime=gethrtime();
	
	// obter numero de nanosegundos desde o inicio da medicao
	res.tm=actTime-localMeasures[*descriptor];
	
	res.ts=actTime;
	res.stat=OK;
	
	rtf_put(fifo,&res,sizeof(res));
	
	//libertar descriptor
	localMeasures[*descriptor]=0;	
}



/*
+--------------------------------------------------------------------------+
|                           EXECUCAO ATEMPADA                              |**
+--------------------------------------------------------------------------+**
   ***************************************************************************
*/


/*
+----------------------------------------------------------------+
| verifica se e executavel; recalcula delay e adormece newdelay; |
| executa a funcao e devolve a aplicacao o instante final        |
+----------------------------------------------------------------+
*/

void* execTask(void * par)
{
	hrtime_t now;
	hrtime_t nanos_past_start_ev;
	hrtime_t newdelay;
	TCBtimestamp res;
	hrtime_t end_ev,a,b;
	int desc=(int)par;	
	struct sched_param p;
	rtl_irqstate_t irqstate;

   	p.sched_priority=TCB_USER_TASK_PRIORITY;    
	
   	pthread_setschedparam(pthread_self(),SCHED_FIFO,&p);	
	
	rtl_no_interrupts(irqstate);		   
		
	//instante actual
	now=gethrtime();
	
	// nanosegundos desde o instante inicial
	nanos_past_start_ev=now-executions[desc].start_ev;
	
	//recalcular o delay
	 if (executions[desc].delay < nanos_past_start_ev )
	   newdelay=0;
	  else
	   newdelay=executions[desc].delay-(nanos_past_start_ev);

	a=now+newdelay+executions[desc].wcet;
	b=executions[desc].start_ev + executions[desc].max_exec;

#ifdef DEBUG
	rtl_printf("Exec:\nStart Ev =");
	printTimeStamp(executions[desc].start_ev);
	rtl_printf("\nDelay =");
	printTimeStamp(executions[desc].delay);
	rtl_printf("\nWcet =");
	printTimeStamp(executions[desc].wcet);
	rtl_printf("\nMax Exec =");
	printTimeStamp(executions[desc].max_exec);
	rtl_printf("\nNewDelay =");
	printTimeStamp(newdelay);
	rtl_printf("\nNow =");
	printTimeStamp(now);
	rtl_printf("\n");
#endif
	
	// verificar se e possivel executar atempadamente
	if ( a > b )
	{
	  rtl_restore_interrupts(irqstate);			   
	  //nao e possivel executar atempadamente
	  res=CANT_RUN_TIMELY;
	  rtf_put(executions[desc].fifo,&res,sizeof(res));		
	  executions[desc].func=0;
	  pthread_delete_np(pthread_self());
	}
	 else
	  {
	     tcb_init_user_task(pthread_self(),executions[desc].delay+executions[desc].start_ev,executions[desc].max_exec,executions[desc].TraceName);
	     rtl_restore_interrupts(irqstate);			   	 			     
	     pthread_wait_np();                                                                          
	     
	   //rtl_no_interrupts(irqstate);		   
#ifdef TRACE_TCB
	   tcb_trace_entry(USER_HANDLER_START,gethrtime(),0,executions[desc].TraceName);	
#endif	 
	   executions[desc].func(); // executa a funcao	
	   end_ev=gethrtime();      //obtem o instante final
#ifdef TRACE_TCB
	   tcb_trace_entry(USER_HANDLER_END,end_ev,0,executions[desc].TraceName);	
#endif	   	   
	   //rtl_restore_interrupts(irqstate);			   
	   rtf_put(executions[desc].fifo,&end_ev,sizeof(end_ev)); //envia a aplicacao				
 	   executions[desc].func=0; //liberta entrada na tabela
	  }
	
	pthread_delete_np(pthread_self());
	while (1) {};
}



/*
+---------------------------------+
| inicia execucao atempadada      | 
+---------------------------------+
| retorna a aplicacao:            |
|          timestamp              |
|	   FUNC_NOT_FOUND	  |	
|	   WCET_FUNC_NOT_FOUND	  |			
|	   NO_AVAIL_EXEC          |
+---------------------------------+
*/
	
void TCBexec(int fifo,execPars* pars)
{	
	handtype     funcptr;
	wcetFunctype wcetptr;
	TCBtimestamp res;
	int a,desc=NO_AVAIL_EXEC;
	int ret;
	
	// procurar descriptor livre
	for (a=0;((a<MAX_EXEC)&&(desc==NO_AVAIL_EXEC));a++)
	{
		if (executions[a].func==0)
			desc=a;
	}
	
	if (desc == NO_AVAIL_EXEC)
	{
		res=NO_AVAIL_EXEC;
		rtf_put(fifo,&res,sizeof(res));		
		return;
	}
	
	//obter endereco da funcao	
	funcptr=get_module_symbol(pars->modName,pars->funcName);
	if (funcptr == 0)
	{
		res=FUNC_NOT_FOUND;
		rtf_put(fifo,&res,sizeof(res));		
		return;
	}
		
	wcetptr=get_module_symbol(pars->modName,pars->wcetName);
	
	
	if (wcetptr == 0)
	{
		res=WCET_FUNC_NOT_FOUND;
		rtf_put(fifo,&res,sizeof(res));		
		return;
	}
	
	// preencher entrada na tabela
	executions[desc].func=funcptr;
	executions[desc].fifo=fifo;
	
	executions[desc].wcet=wcetptr();
	executions[desc].delay=pars->delay;
	executions[desc].max_exec=pars->max_exec;
	executions[desc].start_ev=pars->start_ev;
	strcpy(executions[desc].TraceName,pars->TraceName);
	
	ret=pthread_create (&(executions[desc].task_id), NULL, execTask, (void *)desc);
	if (ret!=0)																			
	  {
	    res=NO_AVAIL_EXEC;
   	    rtf_put(fifo,&res,sizeof(res));
          } 
}


/*
+---------------------------------------------------------------------------+
|                       DETECCAO DE FALHAS LOCAIS                           |**
+---------------------------------------------------------------------------+**
   ****************************************************************************
*/


void* localFailure(void* arg)
{
	hrtime_t now;
	hrtime_t nanos_past_start_ev;
	int res;
	int desc=(int)arg;
	rtl_irqstate_t irqstate;
	struct sched_param p;
#ifdef BENCH	
	unsigned long long tor=localFD[desc].start_ev+localFD[desc].spec;
	long long diff;
	unsigned long long t;
#endif
	p.sched_priority=TCB_USER_TASK_PRIORITY;    
	
   	pthread_setschedparam(pthread_self(),SCHED_FIFO,&p);	
	
	res=desc;
	rtf_put(localFD[desc].fifo,&res,sizeof(res));			   	    	   	   
	
	rtl_no_interrupts(irqstate);		
	
	//instante actual
	now=gethrtime();	
	
	// nanosegundos desde o instante inicial
	nanos_past_start_ev=now-localFD[desc].start_ev;
	
	
	 if (localFD[desc].spec <= nanos_past_start_ev )
	   {
	    rtl_restore_interrupts(irqstate);	
	    // a falha ja ocorreu !
	    rtf_get(localFD[desc].fifo,&res,sizeof(res));	
	    res=ALREADY_FAILED;
	    rtf_put(localFD[desc].fifo,&res,sizeof(res));		
	    localFD[desc].func=0;
	    pthread_delete_np(pthread_self());	    
	   }
	  else
	  {	   
	   tcb_init_user_task(pthread_self(),localFD[desc].start_ev+localFD[desc].spec,localFD[desc].deadline,localFD[desc].TraceName);	   
	   rtl_restore_interrupts(irqstate);	
	   pthread_wait_np();
	   
	   
	   // neste ponto pode ser parada por endTCBlocalFD!!!!

	   //rtl_no_interrupts(irqstate);	
#ifdef BENCH
	t=gethrtime();
	diff=t-tor;
	if (diff>max_release_delay_of_handler_LFD)
		max_release_delay_of_handler_LFD=diff;
#endif	
	   localFD[desc].faulty=1;	   
	   //rtl_restore_interrupts(irqstate);	
	   localFD[desc].func();	   
	   
	  } 
	
	pthread_delete_np(pthread_self());	    
	while (1) {};
}

/*
+----------------------------+
| inicia deteccao local      |
+----------------------------+
| retorna a aplicacao        |
|	   descriptor        | 
|	   FUNC_NOT_FOUND    |		
|	   NO_AVAIL_LOCAL_FD |
|	   ALREADY_FAILED    | 
+----------------------------+	   
*/
	
void TCBstartlocalFD(int fifo,localFDPars* pars)
{	
	handtype     funcptr;
	wcetFunctype wcetptr;
	int	     res,ret;
	int a,desc=NO_AVAIL_LOCAL_FD;
	
	// procurar descriptor livre
	for (a=0;((a<MAX_LOCAL_FD)&&(desc==NO_AVAIL_LOCAL_FD));a++)
	{
		if (localFD[a].func==0)
			desc=a;
	}
	
	if (desc == NO_AVAIL_LOCAL_FD)
	{
		res=NO_AVAIL_LOCAL_FD;
		rtf_put(fifo,&res,sizeof(res));		
		return;
	}
	
	//obter endereco da funcao	
	funcptr=get_module_symbol(pars->modName,pars->funcName);
	
	if (funcptr == 0)
	{
		res=FUNC_NOT_FOUND;
		rtf_put(fifo,&res,sizeof(res));		
		return;
	}
	
	wcetptr=get_module_symbol(pars->modName,pars->wcetName);
	
	
	if (wcetptr == 0)
	{
		res=WCET_FUNC_NOT_FOUND;
		rtf_put(fifo,&res,sizeof(res));		
		return;
	}
		
	// preencher entrada na tabela
 	localFD[desc].func=funcptr;
	localFD[desc].fifo=fifo;
	localFD[desc].start_ev=pars->start_ev;
	localFD[desc].deadline=pars->deadline;
	localFD[desc].spec=pars->spec;	
	localFD[desc].faulty=0;
	
	ret=pthread_create (&(localFD[desc].task_id), NULL, localFailure, (void *)desc);
	
	if (ret!=0)																			
	  {
	    res=NO_AVAIL_LOCAL_FD;
   	    rtf_put(fifo,&res,sizeof(res));
          } 
	
}


/*
+--------------------------------+
| termina medicao local          |
+--------------------------------+
| returna a aplicacao            |  
|           OK                   |
|	    INVALID_LOCALFD_DESC |
+--------------------------------+
*/

void TCBend_localFD(int fifo,int* descriptor)
{
	endLocalFDResult res;
	hrtime_t actTime;	
	rtl_irqstate_t irqstate;
	
	/* descriptor invalido ??? */
	if ( (*descriptor<0) || (*descriptor>=MAX_LOCAL_FD) || (localFD[*descriptor].func==0) ) 
	{
		res.stat=INVALID_LOCALFD_DESC;
		rtf_put(fifo,&res,sizeof(res));
		return;
	}
	
	// parar a deteccao
	rtl_no_interrupts(irqstate);		
	if (!localFD[*descriptor].faulty)   
	{
	 pthread_delete_np(localFD[*descriptor].task_id);
	}
	rtl_restore_interrupts(irqstate);		
	
	actTime=gethrtime();
	
	// obter numero de nanosegundos desde o inicio da medicao
	res.procdelay=actTime-localFD[*descriptor].start_ev;
	
	res.end_ev=actTime;
	res.stat=OK;
	res.faulty=localFD[*descriptor].faulty;
	
	rtf_put(fifo,&res,sizeof(res));
	
	//libertar descriptor
	localFD[*descriptor].func=0;	
}


/*
+------------------------------------+
| inicializa as variaveis do sistema | 
+------------------------------------+ 
*/

void initVariables(void)
{
	unsigned int a;
	
	for (a=0;a<65536;a++)
	   port_to_fifo[a]=-1;	
	
	for (a=0;a<MAX_SEND_TABLE_SIZE;a++)
	 send_info[a].is_free=1;
	 
	for (a=0;a<MAX_RCV_PLOAD_TABLE_SIZE;a++)
	 receive_payload_info[a].is_free=1;
	 
	for (a=0;a<MAX_WAIT_INFO_TABLE_SIZE;a++)
	 wait_infos[a].is_free=1; 
	 
	for (a=0;a<MAX_RCV_TABLE_SIZE;a++)
	 receive_info[a].is_free=1; 
	
	for(a=0;a<MAX_TCB_CONNECTIONS;a++)
		usedFifos[a]=false;
		
	for (a=0;a<MAX_LOCAL_MEASURES;a++)		
	    	localMeasures[a]=0;
	    	
	for (a=0;a<MAX_EXEC;a++)		
	    	executions[a].func=0;    
	    	
	for (a=0;a<MAX_LOCAL_FD;a++)		
	    	localFD[a].func=0;    
	    	
	for (a=0;a<NUMBER_OF_MACHINES;a++)
		number_of_detections[a]=0;
	    	
	seq_numbers=1;    		    		    		
	n_sends_in_this_slot=0;
}


/*
+-----------------------------------------------------+
| Cria os Fifos RT usados pela tcb                    | 
| 						      | 
| RETORNA: 0 se criou correctamente todos os fifos ou | 
|          -1 se falhou na criacao de algum fifo.     |  				
+-----------------------------------------------------+ 
*/

int createFifos(void)
{
	int cnt;
	
	rtl_printf("\n-creating the rt-fifos for tcb/applications comunications\n\n");	

	
	rtl_printf("     .tcb tracing/debug fifo (0) size=%d ",SIZEOF_TRACING_FIFO);
	if (rtf_create(0, SIZEOF_TRACING_FIFO))
	{	
		rtl_printf("->failed\n");	
		rtl_printf("\naborting!!!\nTCB not started\n\n\n");
		return -1;		
	}
	else
	  rtl_printf("->succeded\n");
	

	// usado para requerer conexoes a tcb
	rtl_printf("     .tcb request connection fifo (%d) size=%d ",TCB_REQ_FIFO,SIZEOF_FIFOS);
	if (rtf_create(TCB_REQ_FIFO, SIZEOF_FIFOS))
	{	
		rtl_printf("->failed\n");	
		rtl_printf("\naborting!!!\nTCB not started\n\n\n");
		return -1;		
	}
	else
	  rtl_printf("->succeded\n");	


	  	
	// fifos disponiveis para as ligacoes
	for (cnt=0;cnt<MAX_TCB_CONNECTIONS;cnt++)
	{
		rtl_printf("     .tcb connection number %d, fifo (%d) size=%d ",cnt+1,cnt+TCB_CON_STARTFIFO,SIZEOF_FIFOS);
		if (rtf_create(cnt+TCB_CON_STARTFIFO, SIZEOF_FIFOS))
		{	
			rtl_printf("->failed\n");	
			rtl_printf("\naborting!!!\nTCB not started\n\n\n");
			return -1;		
		}
		else
		  	rtl_printf("->succeded\n");	
	}
	
	for (cnt=0;cnt<MAX_TCB_CONNECTIONS;cnt++)
	{
		rtl_printf("     .tcb connection number %d (WaitInfo channel),fifo (%d) size=%d ",cnt+1,cnt+TCB_CON_STARTFIFO+INC_TO_WAIT_FIFO,SIZEOF_WAIT_INFO_FIFOS);
		if (rtf_create(cnt+TCB_CON_STARTFIFO+INC_TO_WAIT_FIFO, SIZEOF_WAIT_INFO_FIFOS))
		{	
			rtl_printf("->failed\n");	
			rtl_printf("\naborting!!!\nTCB not started\n\n\n");
			return -1;		
		}
		else
		  	rtl_printf("->succeded\n");	
	}
	  
	return 0;
}


/*
+-------------------------------------+
| Elimina os Fifos RT usados pela tcb | 
+-------------------------------------+ 
*/

void destroyFifos(void)
{
	int cnt;

	rtf_destroy(TCB_REQ_FIFO);
	
	for (cnt=0;cnt<MAX_TCB_CONNECTIONS;cnt++)
		{
		 rtf_destroy(cnt+TCB_CON_STARTFIFO);	
		 rtf_destroy(cnt+TCB_CON_STARTFIFO+INC_TO_WAIT_FIFO);	
		}
}

void free_resources(int fifo)
{
	int a;
	rtl_irqstate_t irqstate;
 	
 	//libertar medicoes locais pendentes            
     	for (a=0;a<MAX_EXEC;a++)
     	{
     		if ((executions[a].func!=0)&&(executions[a].fifo==fifo))
     		{
     		 rtl_no_interrupts(irqstate);
		 pthread_delete_np(executions[a].task_id);
		 executions[a].func=0;
 		 rtl_restore_interrupts(irqstate);
     		}
     	}
 	
 	//libertar execucoes atempadas pendentes            
     	for (a=0;a<MAX_EXEC;a++)
     	{
     		if ((executions[a].func!=0)&&(executions[a].fifo==fifo))
     		{
     		 rtl_no_interrupts(irqstate);
		 pthread_delete_np(executions[a].task_id);
		 executions[a].func=0;
 		 rtl_restore_interrupts(irqstate);
     		}
     	}
     
        //libertar detecoes de falhas locais pendentes        
        for (a=0;a<MAX_LOCAL_FD;a++)
        {
        	if ((localFD[a].func!=0)&&(localFD[a].fifo==fifo))
        	{
 		rtl_no_interrupts(irqstate);
 		if (!localFD[a].faulty)   
		 pthread_delete_np(localFD[a].task_id);
		localFD[a].func=0;
 		rtl_restore_interrupts(irqstate);
        	}
        }        
        
        //deteccoes distribuidas em execucao
        //basta limpar a tabela de winfo - evitar que o handler dispare
        
        for (a=0;a<MAX_WAIT_INFO_TABLE_SIZE;a++)
        {
 		rtl_no_interrupts(irqstate);
 		if ( (wait_infos[a].is_free==0)&&(wait_infos[a].owner!=-1)&&(wait_infos[a].owner==fifo))
 		{
 		  if (wait_infos[a].rt_task_id!=0)
 		  {
 		    pthread_delete_np(wait_infos[a].rt_task_id);
 		    wait_infos[a].rt_task_id=0;
 		  }  
 		}
 		rtl_restore_interrupts(irqstate);        	
        }
        
        
        
        
}


/*
+---------------------------------+
| recebe pedidos de servico a tcb | 
+---------------------------------+ 
*/

int serviceHandler(unsigned int fifo)
{
	ReqTcbHeader reqHead;
	char buff[SIZEOF_FIFOS-sizeof(ReqTcbHeader)];
	int code;
	int * p;
	
	if (rtf_get(fifo,&reqHead,sizeof(ReqTcbHeader))==sizeof(ReqTcbHeader))
	 {
	 
	   if (reqHead.sizeofmsg!=0)
	    rtf_get(fifo,&buff,reqHead.sizeofmsg);	
	    
	 	switch (reqHead.srvc)
	 	{
	 		// fechar conexao
	 		case end_connection:
	 				    free_resources(fifo);
	 				    usedFifos[fifo-TCB_CON_STARTFIFO]=false;  				    		
	 				    break;
	 		
	 		// obter estampilha temporal
	 		case get_timestamp:TCBget_timestamp(fifo);
	 				   break;	
	 				   
	 		// obter id da TCB
	 		case get_id: 		TCBget_id(fifo);
	 					break;		   
	 				   
	 		//comecar medicao local
			case start_localM: TCBstart_local_measure(fifo,(RTIME*)&buff);
					   break;
					   
			// terminar medicao local					   
			case end_localM	 : TCBend_local_measure(fifo,(int*)&buff);
					   break;  		 				   
					   
			// execucao atempada		   
			case exec	 : TCBexec(fifo,(execPars*)&buff);
					   break;			   
			
			case start_localFD: TCBstartlocalFD(fifo,(localFDPars*)&buff);					   
					    break;
		        
		        // terminar deteccao local					   
			case end_localFD : TCBend_localFD(fifo,(int*)&buff);
					   break;			   
			
			case start_distributedFD: TCBstart_distributedFD(fifo,(distributedFDPars*)&buff);
					   	 break;
					   	 
			case msg_received_on_payload: msg_rcv_on_payload(fifo,(msg_rcv_on_payloadPars*)&buff);
					   	 break;		   	  			    
			case winfo_to_fifo: p=(unsigned int*)&buff;
					    port_to_fifo[*p]=fifo+INC_TO_WAIT_FIFO;
					    break;
			case stop_winfo_port:p=(unsigned int*)&buff;
					     port_to_fifo[*p]=-1;
					     break;		    		   	 
					
	 	}
	 }
	
	
	return 0;
}



/*
+-----------------------------------------------------+
|  devolve um numero de conexao livre, marca como     | 
|  ocupada, e associa handler que trata os pedidos da | 
|  coneccao criada.                                   | 
| se nao existir devolve NO_AVAIL_CONN                | 	 
+-----------------------------------------------------+ 
*/

int requestfreeConnection(void)
{
	int a;
	
	for (a=0;a<MAX_TCB_CONNECTIONS;a++)
	{
	  if (!usedFifos[a])
	  {
	   usedFifos[a]=true;
	   rtf_create_handler(a+TCB_CON_STARTFIFO, &serviceHandler);
	   return (a+TCB_CON_STARTFIFO);
	  } 
	}
	return NO_AVAIL_CONN;
}



/*
+-----------------------------------------------------+
| handler invocado quando ha actividade no fifo usado | 
| para invocar conexoes a tcb                         | 	 
+-----------------------------------------------------+ 
*/

int conHandler(unsigned int fifo)
{
	ReqTcbHeader reqHead;
	int con;

#ifdef DEBUG
	rtl_printf("Con handler\n");
#endif

	if (rtf_get(TCB_REQ_FIFO,&reqHead,sizeof(ReqTcbHeader))==sizeof(ReqTcbHeader))
	{
		if (reqHead.srvc==get_connection)
		{  
			if (state_of_operation!=NORMAL_STATE)
			{
			 con=IN_INIT_STATE;	
			 rtf_put(TCB_REQ_FIFO,&con,sizeof(int));
			} 
			else
			 {
			  // verifica conexao livre, NO_AVAIL_CONN se nao existe	
			  con=requestfreeConnection();
			  rtf_put(TCB_REQ_FIFO,&con,sizeof(int));
			 } 
		}
	}	

	return 0;
}

int setconHandler(void)
{

	rtl_printf("\n-installing the tcb connection handler on fifo (%d) ",TCB_REQ_FIFO);

	if (rtf_create_handler(TCB_REQ_FIFO, &conHandler))
	{
		rtl_printf("->failed\n");	
		rtl_printf("\naborting!!!\nTCB not started\n\n\n");
		return -1;			
	}
	else
		rtl_printf("->succeded\n");	
		
	return 0;	
		
}

/* *********************************************************************** */



/*
+-----------------------------------------------------------------+
| Verifica se a informacao de falha relativa a uma dada mensagem  |
| ja chegou (no canal de controlo). Processa a analise temporal da|
| da mensagem.							  |			
+-----------------------------------------------------------------+ 
*/

int is_already_in_rcvInfo(msg_rcv_on_payloadPars* pars) 
{
	int a;
	rtl_irqstate_t irqstate;
	int ret=0;
	unsigned long h;
	unsigned long long t;
	
	//rtl_no_interrupts(irqstate);		
	TCBenter_in_critical_section();
	
	for (a=0;a<MAX_RCV_TABLE_SIZE;a++)
	{
		if (receive_info[a].is_free==0)
		 if ((pars->from==receive_info[a].from)&&(pars->seq==receive_info[a].seq))
		 {
		   // ja falhou	
		   if (pars->time>receive_info[a].time_of_failure)	
		    {
		    //rtl_printf("Ja ca tenho e falhou\n");
		    	receive_info[a].fail=1;
		    	receive_info[a].lateness_available=1;
		    	h=pars->time-receive_info[a].time_of_failure;
		    	h=h/1000;
		    	receive_info[a].lateness_degree=(unsigned long)h;		    	
		    }
		     else  // nao falhou 
		      {
		      //rtl_printf("Ja ca tenho e nao falhou\n");
		      	receive_info[a].fail=0;
		      	t=receive_info[a].time_of_failure-pars->time;
		      	if (t<268435456000)  //max: 2^28 us = 2^28*1000 nanos
		      	{
		      	  receive_info[a].lateness_available=1;
		      	  h=(unsigned long)t;
		    	  h=h/1000;
		    	  receive_info[a].lateness_degree=(unsigned long)h;		    	
		    	}
		    	 else
		    	  receive_info[a].lateness_available=0;  
		      }
		    ret=1;	 	
		 }
	}
 	//rtl_restore_interrupts(irqstate);		
 	TCBleave_critical_section();
	return ret;
}

/*
+-----------------------------------------------------------------+
| Esta funcao e chamada pelas aplicacoes quando recebem pacotes   |
| no canal canal de payload.					  | 
+-----------------------------------------------------------------+ 
*/


void msg_rcv_on_payload(int fifo,msg_rcv_on_payloadPars* pars) 
{	
	int c;
	int res;
	unsigned long long act_time;
	rtl_irqstate_t irqstate;
	
	c=0;
	
//rtl_printf("Recebi: FROM:%u SEQ:%u\n",pars->from,pars->seq);
	
	if (!is_already_in_rcvInfo(pars))
	{
	
	//rtl_no_interrupts(irqstate);		
	TCBenter_in_critical_section();
		
	act_time=gethrtime();
	
	while (c<MAX_RCV_PLOAD_TABLE_SIZE)			
	{
	 if (receive_payload_info[c].is_free)
	  break;
	 //Garbage Collector
	 //Quando ocorrem paragens nos nodos da TCB, podem existir mensagens
	 //que sao recebidas no canal de payload, e cuja informacao de falha
	 //nao e recebida no canal de controlo (o nodo emissor falha antes de
	 //enviar a informacao que tem na tabela send_info.) 
	 if (act_time>receive_payload_info[c].clean_up_after)
	  break;
	 c++;
	}
	
	if (c!=MAX_RCV_PLOAD_TABLE_SIZE)
		{
			receive_payload_info[c].is_free=0;
			//valor seguro
			receive_payload_info[c].clean_up_after=pars->time+(2*(SEND_PERIOD+READ_PERIOD));  
			receive_payload_info[c].time_of_receipt=pars->time;
			receive_payload_info[c].seq=pars->seq;
			receive_payload_info[c].from=pars->from;
		        //rtl_printf("??Recebi: FROM:%u SEQ:%u\n",pars->from,pars->seq);
			//rtl_restore_interrupts(irqstate);		
			TCBleave_critical_section();
			res=0;
		  	rtf_put(fifo,&res,sizeof(res));
		}	
		 else   // nao ha espaco na tabela => simular uma omissao no canal de payload
		  {
		        rtl_printf("No Space on rcvpl....omission simulated\n");
		  	//rtl_restore_interrupts(irqstate);		
		  	TCBleave_critical_section();
		  	res=-1;
		  	rtf_put(fifo,&res,sizeof(res));
		  }
      }	  
       else
        {
           res=0;           
	   rtf_put(fifo,&res,sizeof(res));
	   TCBleave_critical_section();
        }		  
}


/*
+-----------------------------------------------------------------+
| Verifica se a mensagem recebida de (from) com a sequencia (seq) | 
| ja foi recebida. Em caso afirmativo, verifica se foi pontual ou |
| atrasada.            						  |	
+-----------------------------------------------------------------+ 
*/

void already_received_on_payload(int a,unsigned char from,unsigned long seq,unsigned long long time_of_failure)
{
	int c;
	unsigned long t;
	unsigned long long l;
	
	for (c=0;c<MAX_RCV_PLOAD_TABLE_SIZE;c++)
	{
		if (receive_payload_info[c].is_free==0)
		{
			// ja foi recebida?	
			if ((receive_payload_info[c].from==from)&&(receive_payload_info[c].seq==seq))
			{//rtl_printf("JA CHEGOU\n");
				// neste caso chegou atrasada
				if (receive_payload_info[c].time_of_receipt>time_of_failure)
				{
					receive_info[a].fail=1;
					receive_info[a].lateness_available=1;
					// calcular o grau de atraso
					t=receive_payload_info[c].time_of_receipt-time_of_failure;
					t=t/1000;
					receive_info[a].lateness_degree=t;		
				}
				 else   // chegou pontualmente
				  {
				  	receive_info[a].fail=0;
					l=time_of_failure-receive_payload_info[c].time_of_receipt;
					//ver se conseguimos colocar o grau de atraso nos 29 bits reservados
				      	if (l<268435456000)  //max: 2^28 us = 2^28*1000 nanos
				      	{
		      	  		  receive_info[a].lateness_available=1;
		      	  		  t=(unsigned long)l;
		    	  		  t=t/1000;
		    	  		  receive_info[a].lateness_degree=t;		    	
		    			}
		    	 		else  // muito grande!!
		    	  		  receive_info[a].lateness_available=0;  			  	
				  }		
			  receive_payload_info[c].is_free=1;  	  
			  return ;
			}
		}
	}
}

inline int get_wait_info_pos(unsigned long seq,unsigned char to)
{
	int c;
	for (c=0;c<MAX_WAIT_INFO_TABLE_SIZE;c++)
	{
	 	if ((wait_infos[c].is_free==0)&&(wait_infos[c].seq==seq)&&(wait_infos[c].from==to))
	 	  return c;
	 	 
	}
	return -1;
}


/*
+-----------------------------------------------------+
|  Envia a informacao relativa a uma dada mensagem    | 
|  as aplicacoes que esperam essa mesma informacao.   | 
|  send_fifo!=-1 significa que o emissor pertence ao  | 
|      ao nodo e tb devera receber a informacao.      | 	 
+-----------------------------------------------------+ 
*/

void reply_wInfo(int pos_in_wi,int send_fifo,unsigned int dest_port)
{
	fail_info_to_app res;

	memcpy(&res.failure_info[0],&wait_infos[pos_in_wi].failure_info[0],sizeof(fail_info)*NUMBER_OF_MACHINES);
	res.from=wait_infos[pos_in_wi].from;
	res.seq=wait_infos[pos_in_wi].seq;
	
	if (send_fifo!=-1)
	  rtf_put(send_fifo,&res,sizeof(res));

	if (port_to_fifo[dest_port]!=-1)
	 rtf_put(port_to_fifo[dest_port],&res,sizeof(res));
	
	if ((wait_infos[pos_in_wi].rt_task_id!=0)&&(wait_infos[pos_in_wi].any_fail==0))
	    pthread_delete_np(wait_infos[pos_in_wi].rt_task_id);
	  
	wait_infos[pos_in_wi].is_free=1;
}

/*
+-----------------------------------------------------+
|  PROCESSA UM PEDIDO DE DETECAO DE FALHA   	      | 
+-----------------------------------------------------+ 
*/

void process_request(unsigned char from,unsigned char to,unsigned long seq,long time_to_failure,unsigned long RndTripVal,unsigned long long time_of_receipt,unsigned int dest_port)
{												
	int a,c;
	unsigned long long time_of_failure;
	char to_array[NUMBER_OF_MACHINES];		
	unsigned char mask=1,dest=to;
	unsigned char number_of_dest=0;
	unsigned int pos_in_wInfo;
		
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	{
		if (mask&dest)
		{
		  to_array[a]=1;
		  number_of_dest++;
		}  
		   else
		    to_array[a]=0;
		mask<<=1;
	}
	
	// nao e destinada a esta tcb entao nao faz nada
	if (!to_array[tcb_machine_id])
	 return;
	
	// coloca na tabela de pedidos recebidos
	for (a=0;a<MAX_RCV_TABLE_SIZE;a++)
	{
		if (receive_info[a].is_free)
		 {
//printk("Pedido %u RND:%u\n",time_to_failure,RndTripVal);
 //rtl_printf("PEDIDO\n");
		 	receive_info[a].is_free=0;
		 	// time_of_recipt contem e a estampilha temporal
		 	// do instante em que a mensagem foi recebida.	
		 	time_of_failure=time_of_receipt+((long long)time_to_failure*1000)-((unsigned long long)RndTripVal*1000);		 			 	
 
//rtl_printf("RNDVAL:%d  TRTF: %d\n",RndTripVal,(long)time_to_failure);  
		 	receive_info[a].time_of_failure=time_of_failure;
//ttf=time_of_failure>>32;	  	
//rtl_printf("TOF: %X %X\n",(unsigned long)ttf,(unsigned long)time_of_failure);
//rtl_printf("ACT: %X %X\n",(unsigned long)gethrtime()>>32,(unsigned long)gethrtime());
		 	receive_info[a].seq=seq;
		 	receive_info[a].from=from;
		 	receive_info[a].to=to;
		 	// enquanto nao se receber atempadamente assume-se	
		 	// que falhou
		 	receive_info[a].fail=1;		 			 	
		 	receive_info[a].destination_port=dest_port;	 	
		 	receive_info[a].lateness_degree=0;
		 	receive_info[a].lateness_available=0;
		 	// verificar se a mensagem ja chegou  
			already_received_on_payload(a,from,seq,receive_info[a].time_of_failure);		 	
			
			// o emissor da mensagem ja criou a estrutura de Winfo
			if (from==tcb_machine_id) return ;
			
			// os destinatarios da mensagem deverao receber a 
			// informacao relativa ao servico de detecao de falhas
			
			for (c=0;c<MAX_WAIT_INFO_TABLE_SIZE;c++)
		 	{
		 		if (wait_infos[c].is_free)
		 		{
		 		 pos_in_wInfo=c;
		 		 break;
		 		} 
		 	}
		 	// tem de se parar o sistema pois ficaria 
		 	// incoerente.
		 	if (pos_in_wInfo==-1)
		 	 {
		 	 	rtl_printf("No space in wInfo process_resp\n");
		 	 	while (1) {};
		 	 }
		 	for (c=0;c<NUMBER_OF_MACHINES;c++)
		 	  wait_infos[pos_in_wInfo].failure_info[c].fail=2;
		 	wait_infos[pos_in_wInfo].is_free=0;
		 	//como recebi um pedido, o emissor da mensagem e remoto 	
		 	wait_infos[pos_in_wInfo].send_fifo=-1; 
		 	wait_infos[pos_in_wInfo].owner=-1;
		 	wait_infos[pos_in_wInfo].n_dests=number_of_dest;  // numero de destinatarios da mensagem
		 	wait_infos[pos_in_wInfo].seq=seq;
		 	wait_infos[a].any_fail=0;
		 	wait_infos[a].rt_task_id=0; // a detecao atempada e no emissor
		 	wait_infos[pos_in_wInfo].to=to;
		 	wait_infos[pos_in_wInfo].from=from;
		 	wait_infos[pos_in_wInfo].destination_port=dest_port;
		 	wait_infos[pos_in_wInfo].failure_at_time=receive_info[a].time_of_failure;
			return;
		 }
	}	
	rtl_printf("ERROR: receive tabled overflow\n");
	while (1) {};   // nao devera acontecer
}


/*
+-----------------------------------------------------+
|  PROCESSA UMA RESPOSTA DE DETECAO DE FALHA   	      | 
+-----------------------------------------------------+ 
*/

void process_resp(unsigned char Bbit,unsigned char to,unsigned char from,unsigned long seq,unsigned long time_to_failure,unsigned int dest_port)
{
	unsigned long h;
	unsigned char fail,lateness_avail;
	unsigned long ldegree;
	int c;
	int pos_in_wInfo;

	h=time_to_failure;
	fail=(h&0x40000000)>>30;
	
	h=time_to_failure;
	lateness_avail=(h&0x20000000)>>29;
	
	h=time_to_failure;
	ldegree=time_to_failure&0x1FFFFFFF;
	
	if (to==tcb_machine_id)  //e para mim
	  number_of_detections[from]--;
	  
	pos_in_wInfo=get_wait_info_pos(seq,to);					
	if (pos_in_wInfo==-1)  //Bbit==1 but is not for me
	{
		//printk("Not for me seq:%u to:%u BBit:%u\n",seq,to,Bbit);
		//for (c=0;c<MAX_WAIT_INFO_TABLE_SIZE;c++)
		//{
		//	printk("SEQ: %u  AV: %u --",wait_infos[c].seq,wait_infos[c].is_free);
		//}
		//while (1) {}
		return;
	}
	//recebi informacao da ocorrencia de uma falha e e a primeira
	if (fail&&!wait_infos[pos_in_wInfo].any_fail)
	{
		//sou o emissor?
		if (wait_infos[pos_in_wInfo].rt_task_id!=0)
		  //acordar a tarefa de deteccao atempada
		  //comeca quando a tarefa de leitura terminar
	  	  pthread_wakeup_np(wait_infos[pos_in_wInfo].rt_task_id);

		wait_infos[pos_in_wInfo].any_fail=1;
	}
	wait_infos[pos_in_wInfo].failure_info[from].fail=fail;
	wait_infos[pos_in_wInfo].failure_info[from].lateness_avail=lateness_avail;
	if (lateness_avail)
	 wait_infos[pos_in_wInfo].failure_info[from].lateness_degree=ldegree;
	  else 
	   wait_infos[pos_in_wInfo].failure_info[from].lateness_degree=0;
	wait_infos[pos_in_wInfo].n_dests--;   
	if (wait_infos[pos_in_wInfo].n_dests==0) //we may reply now because no more info's will came
  	  reply_wInfo(pos_in_wInfo,wait_infos[pos_in_wInfo].send_fifo,dest_port);
		    
	//rtl_printf("Resposta FAIL:<%u> LATE_AVAIL <%u> TO:<%u> FROM<%u> LD<%u>\n",fail,lateness_avail,to,from,ldegree);
}


/*
+-----------------------------------------------------+
| PROCESSA UMA MENSAGEM RECEBIDA NO CANAL DE CONTROLO | 
+-----------------------------------------------------+ 
*/

void process_read(msgtr_hdr *hdr,unsigned long RndTripVal,unsigned long long time_of_receipt)
{
	info_field* pt=(info_field*)(hdr+1);			
	unsigned long h;
	unsigned char R,to;
	unsigned long seq,sign_bit;
	long time_to_failure;
	unsigned char Bbit;
	int items_read=0;
		
	while ((pt->seq_to!=0x00000000)&&(items_read<max_fields)){
	
	 items_read++;		
	 h=pt->r_time_to_failure;
//rtl_printf("R=%u\n",(h&0x80000000)>>31);	
	 R=(h&0x80000000)>>31;   //R=Request bit
	 h=pt->r_time_to_failure;
//printk("time_to_failure=%u\n",(h&0x7FFFFFFF));	
	 if (R==0)         // pedido
	 {
	 sign_bit=(h&0x40000000);        //tempo ate falhar e positivo ou negativo? 	 
	 h=pt->r_time_to_failure;
	 time_to_failure=(h&0x3FFFFFFF);  
	 if (sign_bit)                    // e negativo 
	 {
	  time_to_failure*=-1;
//rtl_printf("MENOS\n");
	 } 
	}
	 else         
	  time_to_failure=(h&0x7FFFFFFF);  
	
	 h=pt->seq_to;
//	 printk("to=%X\n",(h>>24));	
	 to=(h>>24);                     // enderecado a quem?
	 h=pt->seq_to;
//	 printk("seq=%X\n",(h&0x00FFFFFF));	
	 seq=h&0x007FFFFF;		 // sequencia
	
	 h=pt->seq_to;
	 Bbit=((h&0x00800000)!=0);       // B bit activo especifica que
	 				 // todos os nodos devem processar a 
	 				 // a resposta		
		
	
	if (R==0)   // request
	  process_request(hdr->machine_src_id,to,seq,time_to_failure,RndTripVal,time_of_receipt,pt->destination_port);				
	else        // reply
	// so processa as mensagens destinadas a si; como o Bbit activo; ou enviadas por si
	if ((to==tcb_machine_id)||(Bbit)||(hdr->machine_src_id==tcb_machine_id)) 
	{
//rtl_printf("BBIT: %u\n",Bbit);	
	  process_resp(Bbit,to,hdr->machine_src_id,seq,time_to_failure,pt->destination_port);
	}  
	
	pt++;
	}		
}

/*
+-----------------------------------------------------+
| VERIFICA SE NAO RECEBEU INFO DE FALHA DEVIDO A UMA  |
| PARAGEM DE UM NODO.				      | 
+-----------------------------------------------------+ 
*/


void process_wInfos()
{	
	int c,a;
	char to_array[NUMBER_OF_MACHINES];		
	unsigned char mask;
	unsigned char dest;
	unsigned char can_delivery;
	
	for (c=0;c<MAX_WAIT_INFO_TABLE_SIZE;c++)
	{		
		// ja deveriamos ter recebido todas as respostas									
		// este valor e um lower bound - so para efeitos de optimizacao	
		if ((wait_infos[c].is_free==0)&&(gethrtime()>=(wait_infos[c].failure_at_time+SEND_PERIOD+READ_PERIOD))) 
		{
		  mask=1;
		  dest=wait_infos[c].to;
		  
		  for (a=0;a<NUMBER_OF_MACHINES;a++)
		  {
		   if (mask&dest)
  		    to_array[a]=1;
		   else
		    to_array[a]=0;
		    mask<<=1;
		 }
		
		  can_delivery=1;	
		  for (a=0;a<NUMBER_OF_MACHINES;a++)
		  {
		     if (to_array[a])	
		     	//a maquina receptora esta como destinataria mas falhou
		     	//ainda nao recebi mas tb nao vou receber pq a maquina falhou
		  	if ((wait_infos[c].failure_info[a].fail==2)&&(system_view[a]==NOT_PRESENT))
		  	{
		  	 if (wait_infos[c].send_fifo!=-1)		  	 
		  	   number_of_detections[a]--;	
		  	 wait_infos[c].failure_info[a].fail=1;
		  	 wait_infos[c].failure_info[a].lateness_avail=0;
		  	 if (!wait_infos[c].any_fail)
		  	 //se o handler ainda nao foi executado, executar
		  	 //comeca quando a tarefa de leitura terminar
			 {
			   if (wait_infos[c].rt_task_id!=0)
			     {			          
	  	 		  pthread_wakeup_np(wait_infos[c].rt_task_id);
	  	 	     }	  
	 		   wait_infos[c].any_fail=1;
		 	 } 		  	 
		  	}
		  	else
		  	 {
		  	    //ainda podemos receber uma resposta	
		  	    if ((wait_infos[c].failure_info[a].fail==2)&&(system_view[a]==NORMAL))
		  	     can_delivery=0;
		  	 } 
		  }
		  if (can_delivery)		  
  		    reply_wInfo(c,wait_infos[c].send_fifo,wait_infos[c].destination_port);
		}  
	}
}

/*
+--------------------------------------------------------+
|  Constroi o conteudo dos pacotes enviados no rede TR.  |
|  Percorre a tabela send_info, que contem pedidos de    | 
|  detecao pendentes, e constroi os requests especificos.| 
|  Percorre tabela de recive_info: envia info de falhas  |  
+--------------------------------------------------------+ 
*/

int buildDataInfo(msgtr_hdr * msg,unsigned long long ts)
{
	int a,b,c;
	int items=0;
	long time_to_failure;
	unsigned long seq_to=0;
	info_field* pt=(info_field*)(msg+1);
	unsigned long past;
		
	
	//verifica pedidos de detecao pendentes 
	for (a=0;a<MAX_SEND_TABLE_SIZE;a++)
	{
		if (!send_info[a].is_free)
		{
			// quanto tempo passou desde o instante marcado como
			// o inicio da medicao e o evento que assinala 
			// a transmissao da mensagem (do mecanismo de rndtrip)?	
			past=ts-send_info[a].time_of_generation;
			past/=1000;  // em us
			time_to_failure=send_info[a].spec-past;
//rtl_printf("PAST: %u TTF: %d\n",past,time_to_failure);	 
			// ainda nao falhou ?
			if (time_to_failure>=0)
			   time_to_failure=time_to_failure&0x3FFFFFFF;	
			 else // o tempo que falta para falhar e negativo
			  {
			   time_to_failure*=-1;
			   // 31 bits de time_to_failure em complemento 
			   // para dois e negativo
			   time_to_failure=time_to_failure|0x40000000;
			  }
			//Set request bit=0 - significa pedido
			time_to_failure=time_to_failure&0x7FFFFFFF;  
//rtl_printf("TTFFFF:%X\n",time_to_failure);	
			seq_to=send_info[a].to;
			seq_to=seq_to<<24;
		        seq_to=send_info[a].seq|seq_to;
			items++; 
			send_info[a].is_free=1;
			pt->destination_port=send_info[a].destination_port;
			pt->r_time_to_failure=time_to_failure;
			pt->seq_to=seq_to; 
			pt++;									
		}
	}
	
	//receive_info table
	for (a=0;a<MAX_RCV_TABLE_SIZE;a++)
	{
		if (!receive_info[a].is_free)
		{
			if (
			   // tempo expirado  	
			   ((ts>=receive_info[a].time_of_failure)&&(receive_info[a].fail==1))
			   ||
			   //nao falhou enviar a resposta
			   (receive_info[a].fail==0)
			   )
			{
		//printk("RESPOSTA\n");
		//time_to_failure=gethrtime();
		//rtl_printf("TOR: %X %X\n",(unsigned long)(time_to_failure>>32),(unsigned long)gethrtime());
		//time_to_failure=receive_info[a].time_of_failure;
		//rtl_printf("TOR: %X %X\n",(unsigned long)(time_to_failure>>32),(unsigned long)receive_info[a].time_of_failure);
		
			if (receive_info[a].fail==0)
			{
			//rtl_printf("NO FAIL\n");
				if (receive_info[a].lateness_available==1)
				{
				 time_to_failure=receive_info[a].lateness_degree;	 			 
	 			 //request bit =1  no fail bit=0 no lateness_degree bit=1
	 			 time_to_failure=time_to_failure&0x1FFFFFFF;
	 			 time_to_failure=time_to_failure|0xA0000000;
	 			} 
	 			 else
	 			  {
	 			   time_to_failure=0;	 			 
	 			   //request bit =1  no fail bit=0 no lateness_degree bit=0
	 			   time_to_failure=time_to_failure&0x1FFFFFFF;
	 			   time_to_failure=time_to_failure|0x80000000;
	 			  }
			}
			
			if (receive_info[a].fail==1)
			{
			//rtl_printf("FAIL\n");
				if (receive_info[a].lateness_available==1)
				{
				 time_to_failure=receive_info[a].lateness_degree;	 			 
	 			 //request bit =1  fail bit=1 no lateness_degree bit=1
	 			 time_to_failure=time_to_failure|0xE0000000;
	 			} 
	 			 else
	 			  {
	 			   time_to_failure=0;	 			 
	 			   //request bit =1  fail bit=1 no lateness_degree bit=0
	 			   time_to_failure=time_to_failure&0x1FFFFFFF;
	 			   time_to_failure=time_to_failure|0xC0000000;
	 			  }
			}
			
		//	rtl_printf("LD:%u\n",time_to_failure&0x1FFFFFFF);
			
		  	seq_to=receive_info[a].from;
		  	seq_to=seq_to<<24; 
		  	seq_to=receive_info[a].seq|seq_to;
		  	items++; 
		  	pt->r_time_to_failure=time_to_failure;
		  	 pt->seq_to=seq_to; 	
		  	pt->destination_port=receive_info[a].destination_port; 										  	
		  	// B bit
			if (receive_info[a].to!=my_mask)  //set B bit
			 	pt->seq_to=pt->seq_to|0x00800000;
			  else			      // clear B bit
			   	pt->seq_to=pt->seq_to&0xFF7FFFFF;
		  	receive_info[a].is_free=1;
		        pt++;
			}						
		}
	}		
	
	if (items<max_fields)
	{
		pt->seq_to=0x00000000;  //assinala final dos campos
		items++;
		pt++;	
	}
	
	return items;
	
}


void actualize_send_fields(msgtr_hdr * msg,unsigned long diff)
{
	int a;
	info_field* pt=(info_field*)(msg+1);
	long diff_in_micros=diff/1000;
		
	
	//send_info table
	for (a=0;a<MAX_SEND_TABLE_SIZE;a++)
	{
		if (!send_info[a].is_free)
		{
			pt->r_time_to_failure-=diff_in_micros;
			pt++;
		}	
	}	
}

void reinit_infoOf(unsigned char id)
{
	int a;
	
	for (a=0;a<MAX_RCV_PLOAD_TABLE_SIZE;a++)
	{
	 if ((!receive_payload_info[a].is_free)&&(receive_payload_info[a].from==id))
	   receive_payload_info[a].is_free=1;
	} 
	 
	for (a=0;a<MAX_WAIT_INFO_TABLE_SIZE;a++)
	{
	if ((!wait_infos[a].is_free)&&(wait_infos[a].from==id))
	   wait_infos[a].is_free=1;
	} 
	 
	for (a=0;a<MAX_RCV_TABLE_SIZE;a++)
	{
	if ((!receive_info[a].is_free)&&(receive_info[a].from==id))
	   receive_info[a].is_free=1;
	}
	//rtl_printf("RINIT\n");
	system_view[id]=IN_INIT;
	number_of_detections[id]=0;		 	
	number_of_active_machines=get_number_of_normal_state_machines();
}


/*
+--------------------------------------------------------------------------+
|              PROTOCOLO DE DESSIMINACAO PERIODICA E ATEMPADADA            |**
+--------------------------------------------------------------------------+**
   ***************************************************************************
*/


/*
+----------------------------------------------+
| ESTIMA O TEMPO DE ENTREGA DE UMA MENSAGEM    |
+----------------------------------------------+       
*/
unsigned long check_round_trip(msgtr_hdr *msg)  
{
	unsigned long roundtripVal;
	rtl_irqstate_t a;
	char p[2];
 	
 	if (msg->RoundTripInfo[tcb_machine_id].rndval==0)
 	{
 	   rtl_printf("No round trip info in msg\n");
 	   closeReceive();
 	   while (1) {};
 	}
	
 	roundtripVal=(unsigned long)(last_rcv_info[msg->machine_src_id].timestamp-msg->RoundTripInfo[tcb_machine_id].snd_time)/1000;
  	roundtripVal=roundtripVal-msg->RoundTripInfo[tcb_machine_id].rndval;  //in microseconds
  	
   	// como o full-duplex e feito por software, nao se considera o valor
   	// de transmissao minimo quando a mensagem foi enviada e recebida pela
   	// mesma maquina.
  	if (msg->machine_src_id!=tcb_machine_id)  
  	 roundtripVal-=MINIMUM_DELIVERY_DELAY;	

#ifndef BENCH

#ifdef  CONFIG_SELF_CHECKING_CONTROL_CHANNEL_SIMULATED
	if (roundtripVal>MAXIMUM_ROUNDTRIP)
	{
	 rtl_printf("SELF CHECKING value exceded - a value of %u us was estimated \n",roundtripVal);
	 while (1) {};	
	} 
#endif

#ifdef  CONFIG_SELF_CHECKING_CONTROL_CHANNEL_REAL
	if (roundtripVal>MAXIMUM_ROUNDTRIP)
	{
	// shutdown 	
	rtl_no_interrupts(a);
	machine_restart(&p);
	} 
#endif

#endif

#ifdef BENCH
	if (roundtripVal>max_roundtrip)
	  max_roundtrip=roundtripVal;	
#endif

 	return roundtripVal; 	
}

/*
+----------------------------------------------+
| REINICIA O PROTOCOLO DE INICIALIZACAO DO     |
| SISTEMA.				       |
+----------------------------------------------+       
*/

void reinit_protocol(msgtr_hdr *hdr)
{
	int a;
	unsigned char rndcont;

		get_random_bytes(&rndcont,1);   
		if (rndcont<10)
		 rndcont+=10;
#ifdef DEBUGNET
		rtl_printf("Reiniciando o protocolo - cont=%u\n",rndcont);
#endif 		   
		state_of_operation=INIT_STATE;
		contention_rounds=(unsigned char)rndcont;
		for (a=0;a<NUMBER_OF_MACHINES;a++)
		{
		 system_view[a]=NOT_PRESENT;
		 last_rcv_info[a].seq=0;
		 last_rcv_info[a].snd_time=0;
	 	 last_rcv_info[a].timestamp=0;
		}
		system_view[tcb_machine_id]=IN_INIT;
		for (a=0;a<NUMBER_OF_MACHINES;a++)
		  hdr->RoundTripInfo[a].rndval=0;
		number_of_active_machines=0;  
		wait_round=1;	
}

/*
+----------------------------------------------+
| INFORMA O DRIVER DA REDE DE CONTROLO DE QUE  |
| PODE COMECAR A RECEBER OS PACOTES.	       |
+----------------------------------------------+       
*/

void openReceive(void)
{
	rtl_irqstate_t irqstate;
	
	rtl_no_interrupts(irqstate);		   	
   	startReceive();	 
	rtl_restore_interrupts(irqstate);
}


/*
+----------------------------------------------+
| DIZ AO DRIVER DA REDE DE CONTROLO PARA       |
| IGNORAR OS PACOTES RECEBIDOS.		       |
+----------------------------------------------+       
*/

void closeReceive(void)
{
	rtl_irqstate_t irqstate;
	
	rtl_no_interrupts(irqstate);		   	
   	startReceive();	  
	rtl_restore_interrupts(irqstate);
}

/*
+-----------------------------------------------------+
| ENVIA UMA MENSAGEM DE INICIALIZACAO DO PROTOCOLO    |
| DE DESSIMINACAO...NA REDE DE CONTROLO		      |
+-----------------------------------------------------+       
*/

void send_init_message()  
{
	msgtr_hdr *hdrSnd=(msgtr_hdr *)&net_msg_memory;	
	
	hdrSnd->seq=1;	
	hdrSnd->msg_type=INIT_MSG;
	hdrSnd->snd_time=gethrtime();
 	sendMessage(net_msg_memory,sizeof(msgtr_hdr),NETWORK_OMISSION_DEGREE+1); 	
 	hdrSnd->seq++;	
}

/*
+-----------------------------------------------------+
| DETECTOR DE PARAGENS				      |
+-----------------------------------------------------+       
*/

/*
  detectar falha (crash) quando
  
           Penvio    
P  ----|-------------|-----------------------
       | min          \ max
       |               \
Q  ----V---------------->---------|-----------
       a  	 	b Pleitura
 	 
  Q detecta a falha de P no instante b=a+Penvio+(max-min)+l
  
  l=margem de seguranca: devido aos desvios dos relogios e outros factores :-)
                         temos de dar uma margem de seguranca... 
  
  No entanto, visto que funcao e chamada pela tarefa de leitura, a latencia 
  maxima de deteccao e dada por a+Penvio+(max-min)+l+Pleitura.
 	 		 
*/ 	 		 


void process_system_view()
{
	char old_view[NUMBER_OF_MACHINES];  
	int a;

#ifdef DEBUGNET	
memcpy(&old_view,&system_view,NUMBER_OF_MACHINES);   
#endif
	
	number_of_active_machines=0;
	for (a=0;a<NUMBER_OF_MACHINES;a++)    //CRASH detector
	{
	if ((a!=tcb_machine_id)&&(system_view[a]==NORMAL))
	{
	 if (last_rcv_info[a].timestamp<(gethrtime()-(SEND_PERIOD+READ_PERIOD+SECURITY_MARGIN_FOR_NODE_CRASH_DETECTION+((MAXIMUM_ROUNDTRIP*1000)-(MINIMUM_DELIVERY_DELAY*1000)))))
	   system_view[a]=NOT_PRESENT;
	}
	if (system_view[a]==NORMAL)
	  number_of_active_machines++;
	}   

#ifdef DEBUGNET	
if (memcmp(old_view,system_view,NUMBER_OF_MACHINES)!=0)   //DEBUG
 {
 	for (a=0;a<5;a++)
 	 if (system_view[a]==NORMAL)
 	 {
 	  rtl_printf("%u-",a);
 	 }
 	 rtl_printf("\n");
 }
 #endif
		
}


void* ReadTask(void* arg)
{
	struct sched_param p;
	int ret;
	hrtime_t t;
	char indice;
	msgtr_hdr *hdr;
	int a;
	rtl_irqstate_t irqstate;
	pthread_attr_t att;	
	unsigned long rndTrip;
	
	
#ifdef BENCH
	hrtime_t ti,te,tdiff;
#endif	
	
	p.sched_priority=TCB_SYSTEM_TASK_PRIORITY;    
   	pthread_setschedparam(pthread_self(),SCHED_FIFO,&p);	
   	
   	tcb_init_system_task(READ_TASK,pthread_self(),start_of_system_tasks+(SEND_PERIOD/2),READ_PERIOD);
	
	openReceive();	   	
			
	while (1) {
	  	
 	  pthread_wait_np();
	  //rtl_no_interrupts(irqstate);
	  TCBenter_in_critical_section();
#ifdef BENCH
	ti=gethrtime();
	tdiff=ti-next_release_of_read;
	if (tdiff>max_release_time_read)
	 max_release_time_read=tdiff;
	next_release_of_read+=READ_PERIOD; 
#endif 	  
	  memset(&msgfrom_in_this_read_slot,0,NUMBER_OF_MACHINES); 
 	  indice=0;
 	  ret=readMessage((unsigned char*)&(rcv_msg_memory[indice*netMTU]),&t);     
 	  
 	  if ((ret!=0)&&(state_of_operation==SYNC_STATE))
 	  {
 	   state_of_operation=INIT_STATE;
 	   read_loops=0;
 	   send_init_message();
 	   wait_round=1;
 	  } 
 	   
 	  while (ret!=0)
 	  {
 	  	hdr=(msgtr_hdr *)&(rcv_msg_memory[indice*netMTU]); 	  	
 	  
 	  	if ((state_of_operation==INIT_STATE)&&(contention_rounds>0))
 	  	  goto endloop;
 	  	
 	  	if ((hdr->machine_src_id>=NUMBER_OF_MACHINES)||(hdr->machine_src_id<0))
 	  	 {
 	  	  rtl_printf("Error Machine ID Invalid\n");
 	  	  closeReceive();
 	  	  while (1) {};
 	  	 }
 	  	
 		if (((hdr->msg_type==INIT_MSG))&&(state_of_operation!=NORMAL_STATE)&&(hdr->machine_src_id!=tcb_machine_id))  		 
 		{															
 		   //give up and reinits the protocol
#ifdef DEBUGNET 		   
		rtl_printf("Recebi Init em init\n");
#endif		
 		   ret=1;
 		   //flush the message buffer
 		   while (ret!=0)
 		    ret=readMessage((unsigned char*)&(rcv_msg_memory[indice*netMTU]),&t);       
 		   reinit_protocol(hdr);		
 		}
 		else
 	  	{ 
 	  		  	 
 	  	if (hdr->msg_type==INIT_MSG) // flush de seq info from the machine
 	  	 {
 	  	 	  last_rcv_info[hdr->machine_src_id].seq=0;
 	  	 	  
 	  	 	  reinit_infoOf(hdr->machine_src_id);
#ifdef DEBUGNET
		rtl_printf("Recebi init de %d\n",hdr->machine_src_id);
#endif 	  	 	  
 	  	 }  	  	  	  	 
		
		if (last_rcv_info[hdr->machine_src_id].seq==0)
		{
 	  	   	last_rcv_info[hdr->machine_src_id].seq=hdr->seq;
			last_rcv_info[hdr->machine_src_id].timestamp=t;   
			last_rcv_info[hdr->machine_src_id].snd_time=hdr->snd_time;
		}
		else
		{
 	  	 	  	  	  	
 	  	if (hdr->seq==last_rcv_info[hdr->machine_src_id].seq)
 	  	 {
 	  	  last_rcv_info[hdr->machine_src_id].timestamp=t; 	  	  
 	  	 }
 	  	  else
 	  	   if
 	  	   ( 
 	  	    (hdr->seq==last_rcv_info[hdr->machine_src_id].seq+1) ||
 	  	    ((hdr->seq==1)&&(last_rcv_info[hdr->machine_src_id].seq==255))
 	  	   ) 
 	  	   {
 	  	   	if (msgfrom_in_this_read_slot[hdr->machine_src_id])
 	  	   	 {
 	  	   	   rtl_printf("2 Messages from the same machine on the same read slot\n");
 	  	   	   closeReceive();
 	  	   	   while (1)
 	  	   	   {
 	  	   	   };
 	  	   	 }
 	  	   	  else
 	  	   	   msgfrom_in_this_read_slot[hdr->machine_src_id]=1;
 	  	   	
 	  	   	last_rcv_info[hdr->machine_src_id].seq=hdr->seq;
			last_rcv_info[hdr->machine_src_id].timestamp=t;   
			last_rcv_info[hdr->machine_src_id].snd_time=hdr->snd_time;
	 	  	indice++;
	 	  	if (indice==NUMBER_OF_MACHINES+1)
	 	  	   {
	 	  	    rtl_printf("Too many messages in the read slot\n");	 	  	    
	 	  	    closeReceive();
	 	  	    while (1){};
	 	  	   }
	 	  }	
	      	          else
	      	          {				
	 	            rtl_printf("******* MESSAGE LOSTED **********\n");	
	 	            closeReceive();
	 	            while (1) {};
	 	          }  	 	    	 	
	         }
	         }
	         endloop:	         
  	  	 ret=readMessage((unsigned char*)&(rcv_msg_memory[indice*netMTU]),&t);     
 	  } //end While ret
	  
	  if (contention_rounds>0)
	   goto endCont;
	  
	  if (wait_round==0)
	   read_loops++;
	 	 
	 if ((indice==0)&&(read_loops==SYNC_ROUNDS))
	  {
	    wait_round=1;      // I'm the only 
	    state_of_operation=INIT_STATE;
	    send_init_message();
	    read_loops=0;
#ifdef DEBUGNET
	  rtl_printf("No activity on SYNC\n");	
#endif	    
	  }  
	 	 
 	  if (state_of_operation==NORMAL_STATE)	 
   	    process_system_view();
	 
 	  for (ret=0;ret<indice;ret++)
 	   {
 	     hdr=(msgtr_hdr *)&(rcv_msg_memory[ret*netMTU]);
 	  
 	//  rtl_printf("MSG %d SEQ:%d\n",hdr->msg_type,hdr->seq);
 	     
 	     if ((state_of_operation==INIT_STATE)&&(contention_rounds==0))
 	    	system_view[hdr->machine_src_id]=NORMAL;   
 	     else
 	      if ((state_of_operation==NORMAL_STATE)&&(hdr->msg_type!=INIT_MSG))
 	      {  	         	      	   
 	        if (system_view[hdr->machine_src_id]!=NOT_PRESENT)
 	        {   
 	          if (system_view[hdr->machine_src_id]==IN_INIT)
 	             {
 	              system_view[hdr->machine_src_id]=NORMAL;
 	              number_of_active_machines++;
#ifdef DEBUGNET 	              
 	              rtl_printf("NN:%u\n",number_of_active_machines);
#endif
 	             } 
 	          rndTrip=check_round_trip(hdr);
 	          process_read(hdr,rndTrip,last_rcv_info[hdr->machine_src_id].timestamp);
 	         }
 	          else
 	           {
 	            rtl_printf("Message received from a machine not present in the system view\n" );	
 	            while (1) {};
 	           } 
 	      }     
 	     
 	   } // end for
 	   
 	   if (state_of_operation==NORMAL_STATE)
 	   {
 	     //para o caso de ocorrerem paragens no sistema	
	     process_wInfos();		
 	   }
 	    	   
 	   
 	   if ((state_of_operation==INIT_STATE)&&(wait_round==10))
 	   {
 	     state_of_operation=NORMAL_STATE;
 	     system_view[tcb_machine_id]=NORMAL;
 	     wait_round=1;
 	     number_of_active_machines=get_number_of_normal_state_machines();
#ifdef DEBUGNET
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	 if (system_view[a]==NORMAL)
	  rtl_printf("%d-",a);
	rtl_printf("\nNM:%u\n",number_of_active_machines);   	  	
#endif 	      
 	   }
 	   
 	   if ((state_of_operation==INIT_STATE)&&(contention_rounds==0)) wait_round++;
 	   					
 	   endCont:
 	   
 	   if ((state_of_operation==INIT_STATE)&&(contention_rounds>0))
 	   {
	    contention_rounds--;
	    if (contention_rounds==0)
	     send_init_message();
	   } 
 	   
#ifdef BENCH
	te=gethrtime();
	tdiff=te-ti;
	if (tdiff>bench_max_exec_read)
	  bench_max_exec_read=tdiff;
#endif 	 
	//rtl_restore_interrupts(irqstate);
	TCBleave_critical_section();
	}//end while(1)
}

/*
+------------------------------------------------------+
| COLOCA NO PACOTE A INFORMACAO RESPECTIVA AO MECANISMO|
| DE CALCULO DE TEMPOS DE IDA-E-VOLTA DAS MENSAGENS.   |
| DEVOLVE O INSTANTE ASSINALADO PARA O INICIO DA       | 
| TRANSMISSAO.				               |            
+------------------------------------------------------+       
*/


unsigned long long buildRoundTripInfo(msgtr_hdr * msg)
{
	int a,id;
	unsigned short valor;
	hrtime_t now=gethrtime();
	
   	for (a=0;a<NUMBER_OF_MACHINES;a++)
   	if (system_view[a]!=NOT_PRESENT)
   	     {
   	      valor=(unsigned long)(now-last_rcv_info[a].timestamp)/1000;   	      
   	      msg->RoundTripInfo[a].rndval=valor;
   	      msg->RoundTripInfo[a].snd_time=last_rcv_info[a].snd_time;
   	      msg->snd_time=now;
   	     } 
   	return now;  
}


#ifdef BENCH
	int fi=0;
#endif

void* SendTask(void* arg)
{
	struct sched_param p;
	msgtr_hdr *hdr=(msgtr_hdr *)&net_msg_memory;
	rtl_irqstate_t irqstate;
	unsigned long long ts,ts2;
	int n_fields;
#ifdef BENCH
	hrtime_t ti,te,tdiff;
#endif	

	start_of_system_tasks=gethrtime();
	
#ifdef BENCH	
	next_release_of_send=start_of_system_tasks+SEND_PERIOD;
	next_release_of_read=start_of_system_tasks+(SEND_PERIOD/2);
#endif
	
   	p.sched_priority=TCB_SYSTEM_TASK_PRIORITY;    
   	pthread_setschedparam(pthread_self(),SCHED_FIFO,&p);	
   	
   	tcb_init_system_task(SEND_TASK,pthread_self(),start_of_system_tasks,SEND_PERIOD);
   	
	while (1)
	{
 	  pthread_wait_np();
 	  //rtl_no_interrupts(irqstate);		   	
 	  TCBenter_in_critical_section();
 	  
#ifdef BENCH
	ti=gethrtime();
	if (fi!=0)
	{
	  tdiff=ti-next_release_of_send;
	  if (tdiff>max_release_time_send)
	   max_release_time_send=tdiff;
	  next_release_of_send+=SEND_PERIOD; 
	}
	 else
	  fi=1;
#endif	 
	   if (state_of_operation==NORMAL_STATE) 
	    {
	     
	     hdr->msg_type=NORMAL_MSG;
	     ts=gethrtime();
	     n_fields=buildDataInfo(hdr,ts);  	     
	     rtl_no_interrupts(irqstate);		   	
	     ts2=buildRoundTripInfo(hdr);
	     //to provide a better accuracy on the FD		 	     
	     actualize_send_fields(hdr,ts2-ts);
 	     sendMessage(net_msg_memory,sizeof(msgtr_hdr)+(sizeof(info_field)*n_fields),NETWORK_OMISSION_DEGREE+1);     
 	     rtl_restore_interrupts(irqstate); 	  	    	    
	     hdr->seq++;	
	     if (hdr->seq==0)
	      hdr->seq++;
	     n_sends_in_this_slot=0; 
	    }  
#ifdef BENCH
	te=gethrtime();
	tdiff=te-ti;
	if (tdiff>bench_max_exec_send)
	  bench_max_exec_send=tdiff;
#endif	
	//rtl_restore_interrupts(irqstate); 	  	    	
	TCBleave_critical_section();
	}
} 


int createSendReceiveTask(void)
{	
	int ret;
	pthread_attr_t att;
	
	pthread_attr_init(&att);
	att.stack_size=10000;

	rtl_printf("\n-creating send task ");
	ret=pthread_create (&snd_task_id, &att, SendTask, (void *)0);
	
	if (ret!=0)																			
	  {
	    return -1;
          }
          
        rtl_printf("->succeded\n");
        
        rtl_printf("\n-creating read task ");
        
        pthread_attr_init(&att);
	att.stack_size=10000;  
	ret=pthread_create (&rcv_task_id, &att, ReadTask, (void *)0);
	
	if (ret!=0)																			
	  {
	    return -1;
          }
          
        rtl_printf("->succeded\n");  
	return 0;
}

void init_networking(void)
{
	int a;
	msgtr_hdr *hdr=(msgtr_hdr *)&net_msg_memory;
	
	state_of_operation=SYNC_STATE;
	hdr->machine_src_id=get_tcb_id();
	tcb_machine_id=hdr->machine_src_id;
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	{
	 system_view[a]=NOT_PRESENT;
	 last_rcv_info[a].seq=0;
	 last_rcv_info[a].snd_time=0;
	 last_rcv_info[a].timestamp=0;
	}
	system_view[tcb_machine_id]=IN_INIT;
	for (a=0;a<NUMBER_OF_MACHINES;a++)
	  hdr->RoundTripInfo[a].rndval=0;
	wait_round=0;
	read_loops=0;
	contention_rounds=0;
	rtl_printf("\n-TCB ID <%d>\n",hdr->machine_src_id);
	my_mask=1<<(tcb_machine_id);
	number_of_active_machines=0;
	hdr->seq=1;	
}

void get_max_fields()
{
	int max_mtu;
	int net_mtu=getNetMTU();
	
	netMTU=net_mtu;
	
	max_mtu=NET_FIFO_SIZE/NUMBER_OF_MACHINES;
	
	if (max_mtu>net_mtu)
	 max_mtu=net_mtu;
	
	
	max_mtu=max_mtu/(NETWORK_OMISSION_DEGREE+1);	
	
	max_fields=(max_mtu-sizeof(msgtr_hdr))/sizeof(info_field);
		
	max_items=max_fields/(NUMBER_OF_MACHINES+1);
	
	
}


// O escalonador chama esta funcao quando o sistema esta em modo de tracing
void tcb_trace_entry(int evt,unsigned long long time,long arg,char* strarg)
{
	static trace_msg buff;
	
	buff.evt=evt;
	buff.time=time;
	buff.seq=trace_msg_number++;
	buff.arg=arg;
	strcpy(buff.strarg,strarg);
	
	rtf_put(0,&buff,sizeof(buff));		
} 

int init_module(void)
{			

 	// inicializa as variaveis do sistema
	
	initVariables();	
	
	rtl_printf("\nstarting ...\n\n");
	rtl_printf("+-----------------+\n");
	rtl_printf("|  TCB for Linux  |\n");
	rtl_printf("+-----------------+\n");

	// criar os fifos	
	
	if (createFifos())
	 return -1;
	 
	// associar um handler que e invocado quando chegam pedidos a tcb,
	// ou seja quando chega informacao ao fifo 
	
	if (setconHandler())
	 return -1;
	 
	init_networking(); 
	 
	get_max_fields(); 
	 
	if (createSendReceiveTask())
	  return -1; 
	 	 
	rtl_printf("\n *** TCB started and is running ***\n\n");
	
	//notify_sched_of_tcb_trace_entry(&tcb_trace_entry);
			
	return 0;
}

/*
+-------------------------------------+
| elimina tarefas pendentes           | 
+-------------------------------------+ 
*/

void killPendingTasks(void)  
{
	int a;
	rtl_irqstate_t irqstate;
	 
	rtl_no_interrupts(irqstate);		
	 
	   
	for (a=0;a<MAX_EXEC;a++)		
	{
		if (executions[a].func!=0)
		{
		   pthread_delete_np(executions[a].task_id);		   
		}
		
	}
	    	
	for (a=0;a<MAX_LOCAL_FD;a++)		
	{
	if (localFD[a].func!=0)
		{
		   pthread_delete_np(localFD[a].task_id);
		}
	}
	
	for (a=0;a<MAX_WAIT_INFO_TABLE_SIZE;a++)
	{
	if ((!wait_infos[a].is_free)&&(wait_infos[a].rt_task_id!=0))
		{
		   pthread_delete_np(wait_infos[a].rt_task_id);
		}
	}
	
	stopReceive();	  
	
        pthread_delete_np(snd_task_id);
        pthread_delete_np(rcv_task_id);
	rtl_restore_interrupts(irqstate);
	
}


void cleanup_module(void)
{	
		
	killPendingTasks();
	destroyFifos();
	rtl_printf("\n\nTCB execution was terminated\n\n");
#ifdef BENCH		
	rtf_put(0,&bench_max_exec_read,sizeof(long));		
	rtf_put(0,&bench_max_exec_send,sizeof(long));		
	rtf_put(0,&max_roundtrip,sizeof(long));	
	rtf_put(0,&max_release_time_send,sizeof(long));	
	rtf_put(0,&max_release_time_read,sizeof(long));	
	rtf_put(0,&max_release_delay_of_handler_DFD,sizeof(long));	
	rtf_put(0,&max_release_delay_of_handler_LFD,sizeof(long));	
#endif 	 
}
